"""
Created by adam on 3/16/20
"""
__author__ = 'adam'

if __name__ == '__main__':
    pass


class IAnalyzer:
    @staticmethod
    def analyze( content, **kwargs):
        raise NotImplementedError


class NonEmpty(IAnalyzer):

    @staticmethod
    def analyze( content,  **kwargs ):

        if content is not None:
            if isinstance( content, (int, float) ):
                return True

            if len( content ) > 0:
                return True

        return False



"""
Created by adam on 2/23/20
"""
from CanvasHacks.DAOs.sqlite_dao import SqliteDAO
from CanvasHacks.Models.review_association import ReviewAssociation
from CanvasHacks.Models.status_record import ComplexStatusRecord, FeedbackReceivedRecord, InvitationReceivedRecord,\
    StatusRecord
from CanvasHacks.Definitions.activity import Activity
from CanvasHacks.Definitions.unit import Unit
from CanvasHacks.Repositories.interfaces import IRepo

__author__ = 'adam'

from CanvasHacks.Repositories.mixins import StudentWorkMixin
from CanvasHacks.TimeTools import current_utc_timestamp

if __name__ == '__main__':
    pass


class IStatusRepository( StudentWorkMixin, IRepo ):

    def record( self, student, time_to_record=None ):
        """
        Main method called.
        The invitiation repo will record when the student was
        invited to do a review.
        The version which will normally record via
        record_invited. However, can be overriden by other
        repos
        :param student:
        :param time_to_record:
        :return:
        """
        raise NotImplementedError

    def has_received( self, student_id ):
        raise NotImplementedError

    @property
    def previously_received( self ):
        """
        Returns the correct kind of status record for all students
        who have been sent the thing
        :return: list
        """
        raise NotImplementedError

    def get( self, student ):
        """
        Returns the status record for the student or none
        :param student:
        :return:
        """
        raise NotImplementedError


class InvitationStatusRepository( IStatusRepository ):
    """
    This repository handles records of when a student
    was invited to do a particular activity.

    InitialWork: Not called

    Review: Invite sent is recorded on the reviewer once they are sent the content
     of an author's initial work assignment.

    Metareview: Invite sent is recorded on the author once they are sent the feedback
    created in the Review assignment.

    """

    def __init__( self, dao: SqliteDAO, activity: Activity ):
        """
        Create a repository to handle events for a
        particular activity_inviting_to_complete
        """
        self.activity = activity
        self.session = dao.session

    def get( self, student ):
        """
        Returns the record for the student for this unit or none
        if no record has been created yet
        :param student: Student or student id
        :return: FeedbackReceivedRecord or None
        """
        student_id = self._handle_id( student )
        return self.session.query( InvitationReceivedRecord )\
            .filter( InvitationReceivedRecord.activity_id == self.activity.id )\
            .filter( InvitationReceivedRecord.student_id == student_id )\
            .one_or_none()

    def has_received( self, student ):
        """
        Returns true if the student has been invited to
        complete the assignment
        :param student: Student or student id
        :return:
        """
        record = self.get( student )
        return record is not None

    def received_at( self, student ):
        """Returns the timestamp of when the
        student was invited or None if they haven't been
        """
        rec = self.get( student )
        if rec is not None:
            return rec.sent_at
        return None

    @property
    def previously_received( self ):
        """
        Returns InvitationReceivedRecord for all students
        who have been invited to complete the activity
        :return: list
        """
        return self.session\
            .query( InvitationReceivedRecord )\
            .filter( InvitationReceivedRecord.activity_id == self.activity.id )\
            .all()

    @property
    def previously_received_ids( self ):
        """Returns a list of ids of students who have
        already received invitation
        """
        recs = self.previously_received
        if len( recs ) == 0:
            return recs
        return [ r.student_id for r in recs ]

    def record( self, student, time_to_record=None ):
        """
        Records that a student has been invited to
        complete an activity

        # todo Consider bulk commits if slow

        :param student: Student or student id
        :param time_to_record:
        :return:
        """
        student_id = self._handle_id( student )
        if time_to_record is None:
            time_to_record = current_utc_timestamp()

        rec = InvitationReceivedRecord( student_id=student_id, activity_id=self.activity.id, sent_at=time_to_record )

        self.session.add( rec )
        self.session.commit()
        return rec


class FeedbackStatusRepository( IStatusRepository ):
    """
    Handles records of when a student received feedback
    on their work in the activity.

    InitialWork: Feedback sent is recorded once the reviewer turns in the
    review assignment and the feedback is sent to the author of the initial work

    Review: Feedback sent is recorded once the author turns in the metareview assignment
    and the feedback is sent to the person who peer reviewed the initial work

    Metareview: Not called.

    """

    def __init__( self, dao: SqliteDAO, activity: Activity, review_pairings_activity=None ):
        """
         Create a repository to handle events for a
        particular activity

        :param dao:
        :param activity: The activity feedback is sent on
        :param review_pairings_activity: Activity which the review associations are based upon
        """
        self.activity = activity
        self.session = dao.session

        # if None, that means there's no difference between the activities
        # so we can just use the activity
        self.review_pairings_activity = review_pairings_activity if review_pairings_activity is not None else self.activity

    def get( self, student ):
        """
        Returns the record for the student for this activity or none
        if no record has been created yet
        :param student: Student or student id
        :return: FeedbackReceivedRecord or None
        """
        student_id = self._handle_id( student )
        return self.session\
            .query( FeedbackReceivedRecord )\
            .filter( FeedbackReceivedRecord.activity_id == self.activity.id )\
            .filter( FeedbackReceivedRecord.student_id == student_id )\
            .one_or_none()

    def has_received( self, student ):
        """
        Returns true if the student has been sent feedback on
        the assignment
        :param student: Student or student id
        :return:
        """
        record = self.get( student )
        return record is not None

    @property
    def previously_received( self ):
        """
        Returns FeedbackReceivedRecord for all students
        associated with the activity
        :return: list
        """
        return self.session.query( FeedbackReceivedRecord ).filter(
            FeedbackReceivedRecord.activity_id == self.activity.id ).all()

    @property
    def previously_received_ids( self ):
        """Returns a list of ids of students who have
        already received feedback from another student
        """
        recs = self.previously_received
        if len( recs ) == 0:
            return recs
        return [ r.student_id for r in recs ]

    def record( self, student, time_to_record=None ):
        """
        Creates a record that the student was sent
        feedback.
        NB, student id and activity are primary keys so trying to create
        a new record if one already exists should raise an error

        # todo Consider bulk commits if slow

        :param student: Student object or student id
        :param time_to_record:
        :return: FeedbackReceivedRecord
        """
        student_id = self._handle_id( student )
        if time_to_record is None:
            time_to_record = current_utc_timestamp()

        rec = FeedbackReceivedRecord( student_id=student_id, activity_id=self.activity.id, sent_at=time_to_record )

        self.session.add( rec )
        self.session.commit()

        return rec

    def received_at( self, student ):
        """
        Returns the timestamp of when the student was notified or
        None if they haven't been
        :param student:
        :return:
        """
        rec = self.get( student )
        if rec is not None:
            return rec.sent_at
        return None

    @property
    def reviewers_with_authors_sent_feedback( self ):
        """
        Returns the ids of reviewers who have already had their feedback
        sent to the authors
        Used to filter for sending peer review results to authors
        :return:
        """
        notified = self.session\
            .query( ReviewAssociation )\
            .outerjoin( FeedbackReceivedRecord, FeedbackReceivedRecord.student_id == ReviewAssociation.assessee_id )\
            .filter( ReviewAssociation.activity_id == self.review_pairings_activity.id )\
            .filter( FeedbackReceivedRecord.activity_id == self.activity.id )\
            .filter( FeedbackReceivedRecord.student_id.isnot( None ) )\
            .all()
        # ids of reviewers whose authors have been notified
        notified = [ n.assessor_id for n in notified ]
        return notified

    def remove_reviewers_with_notified_authors( self, reviewers ):
        """
        Given a list of reviewers, checks whether the author
        for each reviewer has received their feedback
        and returns the ids of the reviewers who have not
        :param reviewers: list
        :return: list
        """
        notified = self.reviewers_with_authors_sent_feedback
        # print( "n", notified )

        # Make sure we have a list of ids
        reviewers = [ self._handle_id( r ) for r in reviewers ]
        # print( reviewers )

        # Filter out those who have been notified
        reviewers = [ r for r in reviewers if r not in notified ]
        # print( reviewers )

        return reviewers
        # # has_unnotified = [r for r in reviewers if r ]  #  # authors = []  # for student in reviewers:  #     reviewer_id = self._handle_id(student)  #     r = [ra for ra in notified if ra.assessor_id ]  #  #     authors.append()

        #  #     result = self.session \  #         .query(FeedbackReceivedRecord) \  #         .join(ReviewAssociation, FeedbackReceivedRecord.activity_id == ReviewAssociation.activity_id and FeedbackReceivedRecord.student_id == ReviewAssociation.assessee_id).all()  #         # .filter(FeedbackReceivedRecord.activity_id == self.activity.id) \  #         # .filter(ReviewAssociation.activity_id == self.activity.id) \  #         # .filter(ReviewAssociation.assessor_id == reviewer_id) \  #         # .all()

        # result = self.session\  #     .query(FeedbackReceivedRecord)\  #     .outerjoin(ReviewAssociation, FeedbackReceivedRecord.student_id == ReviewAssociation.assessee_id)\  #     .filter(FeedbackReceivedRecord.activity_id == self.activity.id)\  #     .filter(ReviewAssociation.activity_id == self.activity.id)\  #     .filter(ReviewAssociation.assessor_id == reviewer_id)\  #     .one_or_none()  # .filter(FeedbackReceivedRecord.student_id.is_(None))\  # .one_or_none()

        #     if result is not None:  #         result  #         # authors.append( result.student_id )  #  # return authors

        #  #  # td = {'review_association_table' : env.REVIEW_ASSOCIATIONS_TABLE_NAME,  #       'feedback_sent_table': env.FEEDBACK_RECEIVED_STATUS_TABLE_NAME}  # query = """  # SELECT fb.student_id FROM {feedback_sent_table} fb  # INNER JOIN {review_association_table} ra  # ON fb.student_id = ra.assessee_id  # WHERE ra.assessor_id = :student_id  # AND ra.activity_id = :activity_id  # AND fb.activity_id = :activity_id  # """.format(**td)  #  # authors = []  # for student in reviewers:  #     student_id = self._handle_id(student)  #  #     params = {  #         'activity_id' : self.activity.id,  #           'student_id' : student_id  #     }  #  #     stmt = text(query)  #     stmt = stmt.columns( FeedbackReceivedRecord.student_id)  #     result = self.session.query( FeedbackReceivedRecord.student_id)\  #         .from_statement( stmt )\  #         .params( **params )\  #         .one_or_none()  #     if result is not None:  #         authors.append(result)  # return authors


# class MetareviewStatusRepository( IStatusRepository ):
#     """
#     Since the invitation for the metareview needs to
#     record that the invite was sent and that the invitee received
#     feedback, this will handle both
#     """
#
#     def __init__( self, dao: SqliteDAO, unit: Unit ):
#         """
#         Create a repository to handle events for a
#         particular activity
#         """
#         # The activity that we're recording the student as receiving
#         # feedback on
#         self.feedback_activity = unit.initial_work
#         # The activity that we're inviting the student to complete
#         self.invitation_activity = unit.metareview
#
#         self.invite_repo = InvitationStatusRepository( dao, self.invitation_activity )
#         self.feedback_repo = FeedbackStatusRepository( dao, self.feedback_activity )
#
#     def record( self, student, time_to_record=None ):
#         """
#         Create a record that the student was sent their feedback
#         and invited to complete the metareview
#         :param student:
#         :param time_to_record:
#         :return:
#         """
#         self.invite_repo.record( student, time_to_record )
#         self.feedback_repo.record( student, time_to_record )


# ===================== old


class StatusRepository( StudentWorkMixin, IRepo ):
    """
    The status repository tracks when things were sent to
    particular students.

    DEPRECATED. BEING REPLACED AS OF CAN-44 AND CAN-53

    There are two kinds of things that are sent to students:
        Invitations to complete some activity
        Feedback created by another student on that student's work
    """

    def __init__( self, dao: SqliteDAO, activity: Activity ):
        """
        Create a repository to handle events for a
        particular activity_inviting_to_complete
        """
        self.activity = activity
        self.session = dao.session

    def create_record( self, student_or_id ):
        """Creates and returns a record for the student on the activity_inviting_to_complete.
        It does not set the submitted or notified values
        """
        student_id = self._handle_id( student_or_id )
        rec = StatusRecord( student_id=student_id, activity_id=self.activity.id )
        self.session.add( rec )
        # todo Consider bulk commits if slow
        self.session.commit()
        return rec

    @property
    def previously_invited( self ):
        """Returns a list of ids of students who have
        already been notified
        """
        records = self.session.query( StatusRecord ).filter( StatusRecord.activity_id == self.activity.id ).filter(
            StatusRecord.notified.isnot( None ) ).all()
        return [ r.student_id for r in
                 records ]  # record = self.get_record(student_or_id)  # if record is not None:  #     # this will be true if notified is a datetime  #     return record.notified is not None  # # If the record hasn't been created yet, it will  # # return false. That way this check can be run before  # # the process that creates the records  # return False

    @property
    def previously_received_feedback( self ):
        """Returns a list of ids of students who have
        already received feedback from another student
        """
        records = self.session.query( StatusRecord ).filter( StatusRecord.activity_id == self.activity.id ).filter(
            StatusRecord.results.isnot( None ) ).all()
        return [ r.student_id for r in records ]

    def get_record( self, student_or_id ):
        """
        Returns the record for the student for this unit or none
        if no record has been created yet
        :param student_or_id:
        :return:
        """
        student_id = self._handle_id( student_or_id )
        return self.session.query( StatusRecord ).filter( StatusRecord.activity_id == self.activity.id ).filter(
            StatusRecord.student_id == student_id ).one_or_none()

    def get_or_create_record( self, student_or_id ):
        """Keeping the 2 required methods separate in
        case have some use to do them separately
        """
        record = self.get_record( student_or_id )
        if record is None:
            record = self.create_record( student_or_id )
        return record

    def record( self, student, time_to_record=None ):
        """
        Main method called.
        The invitiation repo will record when the student was
        invited to do a review.
        The version which will normally record via
        record_invited. However, can be overriden by other
        repos
        :param student:
        :param time_to_record:
        :return:
        """
        raise NotImplementedError

    def record_invited( self, student, time_to_record=None ):
        """Record that the student was notified that the activity_inviting_to_complete
         is available
        """
        print( "record_invited", student )
        record = self.get_or_create_record( student )
        record.record_opened( time_to_record )
        self.session.commit()

    # def record_submitted( self, student, time_to_record=None):
    #     """Record that the student has submitted
    #     the activity_inviting_to_complete
    #     """
    #     record = self.get_or_create_record(student)
    #     record.record_submission(time_to_record)
    #     self.session.commit()

    def record_sent_feedback( self, student, time_to_record=None ):
        """Records when feedback was sent to this student

        This is only relevant for the metareview since need to record when the
        feedback from this student was sent out to the person who did the peer review.
        NB, notified already represents when they were
        invited to do the metareview unit, the results of which we are now sending
        """
        print( 'record_sent', student )
        record = self.get_or_create_record( student )
        record.record_sent_results( time_to_record )
        self.session.commit()

    def has_received_message( self, student_id ):
        raise NotImplementedError


class ComplexStatusRepository:
    """This likely won't be used. Keeping it until done simplifying"""

    def __init__( self, dao: SqliteDAO, unit: Unit ):
        """
        Create a repository to handle review assignments for a
        particular activity_inviting_to_complete
        """
        self.content_assignment_id = unit.initial_work.id
        self.session = dao.session

    def _handle_id( self, student_or_int ):
        """
        Takes either a student object or the int value of their id
        and returns the id
        :param student_or_int:
        :return: int
        """
        try:
            student_id = int( student_or_int )
        except TypeError:
            student_id = student_or_int.id
        return student_id

    def load( self ):
        """
        Populates self.data with all existing records for the unit
        :return:
        """
        self.data = self.get_unit_records()

    def get_unit_records( self ):
        """
        Returns all existing records for the unit
        :return: list of StatusReord objects
        """
        return self.session.query( ComplexStatusRecord ).filter(
            ComplexStatusRecord.content_assignment_id == self.content_assignment_id ).all()

    def get_student_record( self, student_or_id ):
        """
        Returns the record for the student for this unit or none
        if no record has been created yet
        :param student_or_id:
        :return:
        """
        student_id = self._handle_id( student_or_id )
        return self.session.query( ComplexStatusRecord ).filter(
            ComplexStatusRecord.content_assignment_id == self.content_assignment_id ).filter(
            ComplexStatusRecord.student_id == student_id ).one_or_none()

    def create_record( self, student_or_id ):
        student_id = self._handle_id( student_or_id )
        reviewer_rec = ComplexStatusRecord( student_id=student_id, content_assignment_id=self.content_assignment_id )
        self.session.add( reviewer_rec )
        # commit here?
        return reviewer_rec

    def record_review_assignments( self, list_of_tuples ):
        """
        Expects a list of ( reviewer student, reviewee student) tuples
        :param list_of_tuples:
        :return:
        """
        for reviewer, reviewee in list_of_tuples:
            reviewer_rec = self.get_student_record( reviewer )
            # Get the student. If no record exists, create it
            if reviewer_rec is None:
                reviewer_rec = self.create_record( reviewer )

            # Make sure that a record exists for the reviewee
            reviewee_rec = self.get_student_record( reviewee )
            if reviewee_rec is None:
                reviewer_rec = self.create_record( reviewee )

            reviewer_rec.add_reviewee( reviewee )

        # also store reviewor here?

        # todo commit here or rest_timeout for successful notification?
        self.session.commit()

    def record_peer_review_results_sent( self, list_of_students ):
        """
        For each of the students (id or object), records that the initial
        unit has been sent to the reviewer
        :param list_of_students:
        :return:
        """
        for s in list_of_students:
            record = self.get_student_record( s )
            record.record_sending_review_results()

        self.session.commit()

    def record_metareview_results_sent( self, list_of_students ):
        """
        For each of the students (id or object), records that the results of
        the metareview has been sent to the reviewer
        :param list_of_students:
        :return:
        """
        for s in list_of_students:
            record = self.get_student_record( s )
            record.record_sending_metareview_results()

        self.session.commit()



"""
Created by adam on 1/31/19
"""
# from CanvasHacks.PeerReviewed.Definitions import Unit

__author__ = 'adam'
import configparser
import os
from canvasapi import Canvas

# Needed for managing some test cases
import CanvasHacks.testglobals


CREDENTIALS_FOLDER_PATH = "{}/private"
LIVE_CREDENTIALS = "{}/canvas-credentials.ini"
TEST_CREDENTIALS = "{}/test-credentials.ini"


class Configuration( object ):
    archive_folder = False
    # Which assignments we should be grading
    assignments = [ ]
    # Which discussions we should grading
    discussions = []
    # canvas api object
    course = None
    course_ids = [ ]
    canvas_token = False
    canvas_url_base = False
    # Ids of users, e.g., the teacher to exclude from grading operations
    excluded_users = []
    # Whether in test environment
    is_test = False
    log_folder = False
    # a Unit definition object
    unit = None
    # Number of the unit
    unit_number = None
    test_course_id = None

    @classmethod
    def add_assignment( cls, assignment_id, assignment_name=None ):
        cls.assignments.append( (assignment_id, assignment_name) )
        cls.assignments = list( set( cls.assignments ) )

    @classmethod
    def add_course_id( cls, course_id ):
        cls.course_ids.append( course_id )

    @classmethod
    def add_canvas_token( cls, token ):
        cls.canvas_token = token

    @classmethod
    def add_canvas_url_base( cls, url ):
        cls.canvas_url_base = url

    @classmethod
    def add_discussion( cls, topic_id, name=None ):
        cls.discussions.append( (topic_id, name) )
        cls.discussions = list( set( cls.discussions ) )

    @classmethod
    def add_excluded_user( cls, user_id ):
        cls.excluded_users.append(user_id)
        cls.excluded_users = list(set(cls.excluded_users))

    @classmethod
    def get_assignment_ids( cls ):
        return [ i[ 0 ] for i in cls.assignments ]

    @classmethod
    def get_discussion_ids( cls ):
        return [ i[ 0 ] for i in cls.discussions ]

    @classmethod
    def remove_assignment( cls, assignment_id ):
        el = list( filter( lambda x: x[ 0 ] == assignment_id, cls.assignments ) )[ 0 ]
        idx = cls.assignments.index( el )
        return cls.assignments.pop( idx )

    @classmethod
    def remove_discussion( cls, topic_id ):
        el = list( filter( lambda x: x[ 0 ] == topic_id, cls.discussions ) )[ 0 ]
        idx = cls.discussions.index( el )
        return cls.discussions.pop( idx )

    @classmethod
    def reset_assignments( cls ):
        cls.assignments = [ ]
        print( "List of assignments is empty" )

    @classmethod
    def reset_course_ids( cls ):
        cls.course_ids = [ ]
        print( "List of course ids is empty" )

    @classmethod
    def reset_canvas_token( cls ):
        cls.canvas_token = False
        print( "Canvas token reset to empty" )

    @classmethod
    def reset_config( cls ):
        cls.reset_canvas_token()
        cls.reset_course_ids()

    # ------------ Unit selection
    @classmethod
    def set_unit_number( cls, unit_number, name=None ):
        cls.unit_number = unit_number

    @classmethod
    def initialize_canvas_objs( cls ):
        """Sets a canvas, course, and unit object on the config
        Creates a unit object by downloading from canvas using the first defined course id
        todo Make this work with multiple courses
        """

        COURSE_ID = cls.course_ids[0]
        # print("Working on course: ", COURSE_ID)
        cls.canvas = Canvas(cls.canvas_url_base, cls.canvas_token)
        cls.course = cls.canvas.get_course(COURSE_ID)
        # if cls.unit_number is not None:
        #     cls.unit = Unit(cls.course, cls.unit_number)

    @classmethod
    def reset_unit_number( cls, dummy_param=None):
        cls.unit = None
        cls.unit_number = None

    @classmethod
    def get_unit_number( cls ):
        """I know, stupid. But it parallels other methods
        so the selection buttons will work the same"""
        return [cls.unit_number]

    @classmethod
    def load_section_ids( cls ):
        try:
            for v in cls.configuration[ 'sections' ].values():
                cls.add_course_id( int(v) )
        except:
            pass

    # --------------- Test vs live
    @classmethod
    def set_test( cls ):
        """Tells to set all variables to test settings"""
        cls.is_test = True
        # set course id
        test_id = cls.configuration['testing'].get('TEST_COURSE_ID')
        cls.course_ids = [int(test_id)]
        print(" ".join([" TEST " for _ in range(0, 5)]))
        # Set on the global variable (which is only used in certain tests)
        CanvasHacks.testglobals.TEST = True

    @classmethod
    def set_live( cls ):
        """Set all variables to live setting"""
        cls.is_test = False
        # clear out possible test values
        cls.course_ids = []
        # Set to the stored course ids
        cls.load_section_ids()
        print(" ".join([" LIVE " for _ in range(0, 5)]))
        # Set on the global variable (which is only used in certain tests)
        CanvasHacks.testglobals.TEST = False


class InteractiveConfiguration( Configuration ):
    def __init__( self ):
        super().__init__()

    @classmethod
    def handle_token_entry( cls, event ):
        if event[ 'type' ] == 'change' and event[ 'name' ] == 'value':
            v = event[ 'new' ]
            cls.add_canvas_token( v )

    @classmethod
    def handle_url_entry( cls, event ):
        if event[ 'type' ] == 'change' and event[ 'name' ] == 'value':
            v = event[ 'new' ]
            cls.add_canvas_url_base( v )

    @classmethod
    def set_test( cls ):
        cls.is_test = True

    @classmethod
    def set_live( cls ):
        cls.is_test = False


class FileBasedConfiguration( Configuration ):
    configuration = False
    file_path = False

    def __init__( self, configuration_file_path ):
        super().__init__()
        FileBasedConfiguration.file_path = configuration_file_path

    @classmethod
    def read_config_file( cls ):
        if not cls.file_path:
            # The file path could've been customized by instantiating the class
            # If that didn't happen, we go with the default
            # The folder containing the assets folder
            cls.proj_base = os.path.abspath( os.path.dirname( os.path.dirname( __file__ ) ) )
            # All login credentials are defined in files here.
            # THIS CONTENTS OF THIS FOLDER MUST NOT BE COMMITTED TO VERSION CONTROL!
            folder_path = CREDENTIALS_FOLDER_PATH.format(cls.proj_base)
            creds = TEST_CREDENTIALS if cls.is_test else LIVE_CREDENTIALS
            cls.file_path = creds.format(folder_path )

        cls.configuration = configparser.ConfigParser()
        cls.configuration.read( cls.file_path )
        print( "Reading credentials and settings from %s" % cls.file_path )

    @classmethod
    def load( cls, is_test=False ):
        cls.is_test = is_test
        cls.read_config_file()
        cls.load_token()
        cls.load_url_base()
        cls.load_local_filepaths()
        cls.load_section_ids()
        cls.load_excluded_users()
        cls.initialize_canvas_objs()
        cls.semester_name = cls.configuration['names'].get('SEMESTER')

    @classmethod
    def load_local_filepaths( cls ):
        root = os.getenv( "HOME" )
        cls.archive_folder = "{}/{}".format(root, cls.configuration[ 'folders' ].get( 'STUDENT_WORK_ARCHIVE_FOLDER' ))
        cls.log_folder = "{}/{}".format(root, cls.configuration[ 'folders' ].get( 'LOG_FOLDER' ))
        cls.data_folder = '{}/data'.format(cls.proj_base)

    @classmethod
    def load_token( cls ):
        if not cls.configuration:
            cls.read_config_file()

        cls.add_canvas_token( cls.configuration[ 'credentials' ].get( 'TOKEN' ) )

    @classmethod
    def load_url_base( cls ):
        cls.add_canvas_url_base( cls.configuration[ 'url' ].get( 'BASE' ) )

    @classmethod
    def load_excluded_users( cls ):
        try:
            for v in cls.configuration[ 'excluded_users' ].values():
                cls.add_excluded_user(int(v))
            print("Will ignore work by users: ", cls.excluded_users)
        except:
            pass


class TestingConfiguration( Configuration ):
    def __init__( self ):
        type(self).is_test = True

        super().__init__()

    @classmethod
    def handle_token_entry( cls, event ):
        if event[ 'type' ] == 'change' and event[ 'name' ] == 'value':
            v = event[ 'new' ]
            cls.add_canvas_token( v )

    @classmethod
    def handle_url_entry( cls, event ):
        if event[ 'type' ] == 'change' and event[ 'name' ] == 'value':
            v = event[ 'new' ]
            cls.add_canvas_url_base( v )

    @classmethod
    def set_test( cls ):
        cls.is_test = True

    @classmethod
    def set_live( cls ):
        cls.is_test = False



if __name__ == '__main__':
    FileBasedConfiguration.load()
    print( FileBasedConfiguration.canvas_token )



"""
Created by adam on 2/26/20
"""
__author__ = 'adam'

from CanvasHacks.Errors.data_ingestion import NoStudentWorkDataLoaded, NoNewSubmissions
from CanvasHacks.Definitions.skaa import Review
from CanvasHacks.Repositories.assignments import AssignmentRepository
from CanvasHacks.Repositories.discussions import DiscussionRepository
from CanvasHacks.Repositories.quizzes import ReviewRepository, QuizRepository
from CanvasHacks.Repositories.submissions import QuizSubmissionRepository
from CanvasHacks.Loaders.factories import LoaderFactory

if __name__ == '__main__':
    pass


class WorkRepositoryFactory:
    """Decides what kind of repository is needed
    and instantiates it

    Used for ordinary quiz assignments that aren't part of skaa"""

    @staticmethod
    def make( activity, course=None, only_new=False, **kwargs ):
        # Get the object which will handle loading data
        # if only_new:
        #     loader = NewQuizReportFileLoader( activity, course )
        # else:
        #     loader = AllQuizReportFileLoader( activity, course )

        # Get quiz submission objects
        if isinstance( activity, Review ):
            repo = ReviewRepository( activity, course )
        else:
            repo = QuizRepository( activity, course )
        return repo


class WorkRepositoryLoaderFactory:
    """Decides what kind of repository is needed
    and instantiates it after loading data into it

    Replaces make_quiz_repo
    """

    @classmethod
    def make( cls, activity, course=None, only_new=False, **kwargs ):
        """
        Returns a IContentRepository of the appropriate sort with all the
        relevant data loaded, processed, and stored as a dataframe on data
        kwags may contain
            download=True

        :param activity:
        :param course:
        :param only_new:
        :param kwargs:
        :return: IContentRepository
        """
        # Get the object which handles loading student data
        # The params will determine whether this is from file or download
        loader = LoaderFactory.make( is_quiz=activity.is_quiz_type, is_discussion=activity.is_discussion_type, combo=True, **kwargs)

        if activity.is_discussion_type:
            return cls._for_discussion_type_activity(activity, course, loader, **kwargs)

        if activity.is_quiz_type:
            # If uses a quiz report, downloads report and populates the
            # appropriate repository (QuizRepository or ReviewRepository)
            return cls._for_quiz_type_activity(activity, course, loader, **kwargs)

        return cls._for_assignment_type_activity(activity, course, loader, **kwargs)

    @classmethod
    def _for_quiz_type_activity(cls, activity, course, loader, **kwargs):
        """
        Handles the loading if the activity_inviting_to_complete uses a quiz report
        :param activity:
        :param course:
        :param loader:
        :param kwargs:
        :return:
        """
        # Get quiz submission objects
        if isinstance( activity, Review ):
            repo = ReviewRepository( activity, course )
        else:
            repo = QuizRepository( activity, course )

        try:
            student_work_frame = loader.load( activity, course,  **kwargs )

        # Doing each separately so can modify if needed
        except NoStudentWorkDataLoaded:
            raise NoStudentWorkDataLoaded

        except NoNewSubmissions:
            raise NoNewSubmissions

        if student_work_frame is None or len(student_work_frame) == 0:
            # This indicates that all attempts to acquire data
            # have failed and no further will be attempted
            raise NoStudentWorkDataLoaded

        # Download submissions
        subRepo = QuizSubmissionRepository( repo.quiz )

        # Doing the combination with submissions after saving to avoid
        # mismatches of new and old data
        repo.process( student_work_frame, subRepo.frame )

        return repo

    @classmethod
    def _for_assignment_type_activity(cls, activity, course, loader, **kwargs):
        """
        Handles the loading if the activity_inviting_to_complete uses an unit (with no report)
        :param activity:
        :param course:
        :param loader:
        :param kwargs:
        :return:
        """
        repo = AssignmentRepository( activity, course )

        student_work_frame = loader.load( activity, course, **kwargs )

        if student_work_frame is None or len(student_work_frame) == 0:
            raise NoStudentWorkDataLoaded

        repo.process( student_work_frame )

        return repo

    @classmethod
    def _for_discussion_type_activity( cls, activity, course, loader, **kwargs ):
        repo = DiscussionRepository(activity, course)
        # todo refactor the downloading stuff out of the repository and into the loader
        repo.download()
        #
        # student_work_frame = loader.load( activity, course, **kwargs )
        #
        if len(repo.data) == 0:
            raise NoStudentWorkDataLoaded
        #
        # repo.process( student_work_frame )

        return repo



"""
Created by adam on 3/15/20
"""
__author__ = 'adam'

from unittest import TestCase

from CanvasHacks.GradingMethods.wordcount import GradeByWordCount, make_pct_required_count_thresholds
from TestingBase import TestingBase
import random
if __name__ == '__main__':
    pass

class TestMakePctRequiredCountThresholds(TestingBase):
    def setUp(self) -> None:
        self.config_for_test()

    def test_by_tens( self ):
        required_cnt = 100
        result = make_pct_required_count_thresholds(required_cnt)

        self.assertEqual(10, len(result))

        min_counts = [p for p in range( 0, 100, 10 )]

        for r in result:
            self.assertIn(r['min_count'], min_counts, "Not an unexpected min count")


class TestGradeByWordCount( TestingBase ):

    def setUp(self) -> None:
        self.config_for_test()
        self.dicts = [{ 'min_count': 100, 'pct_credit': 10},
                              { 'min_count': 200, 'pct_credit': 20},
                              { 'min_count': 300, 'pct_credit': 30},
                              { 'min_count': 400, 'pct_credit': 40}]


    def test_prepare_dicts( self ):

        input_dicts = self.dicts
        random.shuffle(input_dicts)
        # self.assertEqual(dicts[0]['count'], 100)

        obj = GradeByWordCount(input_dicts)
        obj.prepare_dicts()

        self.assertEqual(obj.threshold_dicts[0]['min_count'], 400)

    def test_grade( self ):
        # ( count, credit )
        expect = [ ( 50,  0), (100, 10), (275, 20), (888, 40) ]

        obj = GradeByWordCount(self.dicts)

        for count, credit in expect:
            content = " ".join( [ self.fake.word() for _ in range( 0, count ) ] )

            self.assertEqual(obj.grade(content), credit, "Returns correct pct as 0-100")



"""
Created by adam on 9/14/18
"""
__author__ = 'adam'

import os

import CanvasHacks.testglobals

if CanvasHacks.testglobals.TEST:
    print( "RUNNING IN TEST MODE" )

# Putting this here for an easy place to manually
# toggle, since it only needs to be true when
# testing against server.
CanvasHacks.testglobals.TEST_WITH_FILE_DB = False

from CanvasHacks.Configuration import FileBasedConfiguration, InteractiveConfiguration, TestingConfiguration

ROOT = os.getenv( "HOME" )

if CanvasHacks.testglobals.use_api:
    # Check whether it is being run on my machine or remotely
    if ROOT[ :12 ] == '/Users/adam':
        FileBasedConfiguration.load( CanvasHacks.testglobals.TEST )
        CONFIG = FileBasedConfiguration
        TEMP_DATA_PATH = "%s/temp" % FileBasedConfiguration.proj_base
        ARCHIVE_FOLDER = FileBasedConfiguration.archive_folder
        JOURNAL_ARCHIVE_FOLDER = "%s/Journals" % ARCHIVE_FOLDER
        LOG_FOLDER = FileBasedConfiguration.log_folder
        DATA_FOLDER = "%s/data" % FileBasedConfiguration.proj_base
        DOWNLOAD_FOLDER = "{}/Downloads"
        # Testing
        TEST_DATA_PATH = "{}/tests/testdata".format( FileBasedConfiguration.proj_base )

        TOKEN = FileBasedConfiguration.canvas_token
        URL_BASE = FileBasedConfiguration.canvas_url_base

    else:
        CONFIG = InteractiveConfiguration
        TOKEN = InteractiveConfiguration.canvas_token
        URL_BASE = InteractiveConfiguration.canvas_url_base
        # Logging should stream
        LOG_FOLDER = None
else:
    # Testing without api access
    CONFIG = TestingConfiguration
    if CanvasHacks.testglobals.TEST:
        CONFIG.is_test = CanvasHacks.testglobals.TEST
    LOG_FOLDER = None

# DB
REVIEW_ASSOCIATIONS_TABLE_NAME = "review_associations"
STATUS_TABLE_NAME = 'status'
COMPLEX_STATUS_TABLE_NAME = 'complexstatus'
FEEDBACK_RECEIVED_STATUS_TABLE_NAME = 'feedback_received'
INVITATION_RECEIVED_STATUS_TABLE_NAME = 'invitation_received'
SUBMISSION_TABLE_NAME = 'submissions'

STUDENT_TABLE_NAME = 'students'

# Logging
t = "TEST-" if CONFIG.is_test else ""
STUDENT_WORK_PROCESSING_LOGNAME = '{}student-work-processing-log.txt'.format( t )
MESSAGE_LOGNAME = "{}message-log.txt".format( t )
RUN_LOGNAME = "{}run-log.txt".format( t )

# Plotting stuff
LIKERT_PLOT_ORDER = [ 'Forgot', 'Strongly disagree', 'Disagree', 'Agree', 'Strongly agree' ]
LIKERT_NUM_MAP = { 'Forgot': 0, 'Strongly disagree': 1, 'Disagree': 2, 'Agree': 3, 'Strongly agree': 4 }

# File stuff
REPORT_KEEP_COLUMNS = [ 'attempt', 'course_id', 'finished_at_date',
                        'fudge_points', 'id', 'name', 'quiz_id', 'score',
                        'section_sis_id', 'student_id', 'submission_id', 'user_id', 'workflow_state' ]



"""
Defines types of activity_inviting_to_complete with respect to features of
the text we'd expect
Created by adam on 2/25/20
"""
__author__ = 'adam'

class ITextType:
    """
    Parent
    """
    pass


class AuthorActivity(ITextType):
    """
    An activity_inviting_to_complete where they are to discuss an author
    or authors
    """

    def __init__(self, authors):
        authors = authors if isinstance(authors, list) else [authors]
        self.authors = authors
        self.expected += self.authors





if __name__ == '__main__':
    pass


"""
Created by adam on 3/11/20
"""
__author__ = 'adam'

from bs4 import BeautifulSoup

from CanvasHacks.Models.model import StoreMixin
import re

if __name__ == '__main__':
    pass


class ICleaner:
    @staticmethod
    def clean( content ):
        raise NotImplementedError


class UtfCleaner( ICleaner ):

    @staticmethod
    def clean( content ):
        single_char_to_remove = [
            '°',
            'ç',
            'ï',
            '¬', '†',
            "Ä", "ô"  # , "s"
        ]
        multi_char_to_remove = [
            "\xa0"
        ]
        chars = single_char_to_remove + multi_char_to_remove
        rx = r'{}'.format("|".join(chars))
        content = re.sub( rx, '', content )

        # Dealing with single first to reduce time for next step
        # content = "".join( [ c for c in content if c not in single_char_to_remove ] )
        # return str( content )  # .encode('utf-8').decode()
        #
        # for c in multi_char_to_remove:
        #     content = re.sub(c, '', content)
        return content


class HtmlCleaner( ICleaner ):

    @staticmethod
    def clean( content ):
        soup = BeautifulSoup( content, features="lxml" )
        return soup.text


class TextCleaner( StoreMixin ):
    cleaners = [
        UtfCleaner,
        HtmlCleaner
    ]

    def __init__( self, **kwargs ):
        self.handle_kwargs( **kwargs )

    def clean( self, content ):
        if not isinstance(content, str):
            # This is a text cleaner. If we haven't received
            # text, we just send it back
            return content

        try:
            for cleaner in self.cleaners:
                content = cleaner.clean( content )
            return content
        except Exception as e:
            print( 'Text Cleaner Error', e, content )



"""
Created by adam on 3/13/20
"""
__author__ = 'adam'

from CanvasHacks.Repositories.overview import SkaaOverviewRepository

if __name__ == '__main__':
    pass

SKAA_ORDER = [ 'student', 'reviewing', 'reviewed_by', 'invited_to_review',
               'received_feedback_on_essay', 'invited_to_metareview',
               'received_feedback_on_review', 'canvas_id', 'csun_id', 'reviewing_id', 'reviewed_by_id' ]

DISCUSSION_ORDER = [ 'student', 'reviewing', 'reviewed_by', 'invited_to_discussion_review',
                     'received_discussion_feedback', 'canvas_id', 'csun_id', 'reviewing_id', 'reviewed_by_id' ]


class SkaaDashboard:
    """
    This is in charge of displaying information from the
    skaa overview repo
    """

    def __init__( self, overview_repo: SkaaOverviewRepository ):
        self.repo = overview_repo

    @property
    def data( self ):
        return self.repo.data[ SKAA_ORDER ]

    @property
    def essay( self ):
        """
        Return students who have done initial work and been assigned a reviewer

        :return: DataFrame
        """
        return self.repo.essay

    @property
    def no_essay( self ):
        """
        Students who have not submitted the initial work
        :return: DataFrame
        """
        return self.repo.no_essay

    @property
    def reviewed( self ):
        """
        Returns the subset of students who have turned in the initial work
        whose reviewer has turned in the review

        :return: DataFrame
        """
        return self.repo.reviewed

    @property
    def non_reviewed( self ):
        """
        Returns the subset of students who have turned in the initial work
        whose reviewer has NOT turned in the review

        Drops the 'reviewing' column in results since that can be
        confusing in this context
        :return: DataFrame
        """
        return self.repo.non_reviewed.drop( [ 'reviewing' ], axis=1 )

    @property
    def metareviewed( self ):
        """
        Returns the subset of students who have turned in the initial work
        whose author has turned in the metareview

        Drops the 'reviewed_by' column in results since that can be
        confusing in this context

        :return: DataFrame
        """
        return self.repo.metareviewed.drop( [ 'reviewed_by' ], axis=1 )

    @property
    def non_metareviewed( self ):
        """
        Returns the subset of students who have turned in the initial work
        whose author has turned in the metareview

        Drops the 'reviewed_by' column in results since that can be
        confusing in this context

        :return: DataFrame
        """
        return self.repo.non_metareviewed.drop( [ 'reviewed_by' ], axis=1 )

    def print_counts( self ):
        print( "\n~~~~~~~~~~~~~~~~~~~~~ SKAA ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~" )
        print( "===================== initial work =====================" )
        print( "{} students have turned in essay and been paired up".format( len( self.essay ) ) )
        print( "{} students haven't turned in essay".format( len( self.no_essay ) ) )

        print( "\n===================== review =====================" )
        print( "{} students' reviewers has turned in the review".format( len( self.reviewed ) ) )
        print( "{} students have a reviewer who hasn't turned in the review".format( len( self.non_reviewed ) ) )

        print( "\n===================== metareview =====================" )
        print( "{} students' authors have turned in the metareview".format( len( self.metareviewed ) ) )
        print( "{} students' authors haven't turned in the metareview".format( len( self.non_metareviewed ) ) )
        print( "\n" )


class DiscussionDashboard:
    """
    This is in charge of displaying information from the
    discussion overview repository

    """

    def __init__( self, overview_repo ):
        self.repo = overview_repo

    @property
    def data( self ):
        return self.repo.data[ DISCUSSION_ORDER ]

    @property
    def posters( self ):
        """
        Students who have posted and been assigned reviewers
        :return: DataFrame
        """
        return self.repo.posters

    @property
    def non_posters( self ):
        """
        Students who have not posted and thus not been assigned a reviewer
        :return: DataFrame
        """
        return self.repo.non_posters

    @property
    def reviewed( self ):
        return self.repo.reviewed

    @property
    def non_reviewed( self ):
        """
        Returns students whose reviewer hasn't turned in the review.

        Drops the reviewing column to avoid confusion
        :return:
        """
        return self.repo.non_reviewed.drop( [ 'reviewing' ], axis=1 )

    def print_counts( self ):
        print( "\n~~~~~~~~~~~~~~~~~~~ DISCUSSION ~~~~~~~~~~~~~~~~~~~~~~~~~" )
        print( "===================== Discussion posts =====================" )
        print( "{} students have turned in posts and been paired up".format( len( self.posters ) ) )
        print( "{} students have not posted".format( len( self.non_posters ) ) )

        print( "\n===================== review =====================" )
        print( "{} students' reviewers has turned in the review".format( len( self.reviewed ) ) )
        print( "{} students have a reviewer who hasn't turned in the review".format( len( self.non_reviewed ) ) )
        print( "\n" )



"""
Created by adam on 2/26/20
"""
__author__ = 'adam'

from tests.TestingBase import TestingBase

from faker import Faker

fake = Faker()


class TestWorkRepositoryLoaderFactory( TestingBase ):

    def setUp(self):
        self.config_for_test()

    def test_make( self ):
        self.skipTest('todo')

    def test__for_assignment_type_activity( self ):

        self.skipTest('todo')
    #
    # @patch('CanvasHacks.Repositories.factories.WorkRepositoryLoaderFactory.QuizSubmissionRepository')
    def test__for_quiz_type_activity( self ):
        self.skipTest( 'todo' )

        # loader = MagicMock(return_value=pd.DataFrame())
        # WorkRepositoryLoaderFactory
        # subRepoMock.frame = pd.DataFrame()
        #
        # # check
        # subRepoMock.frame.assert_called()



class TestWorkRepositoryFactory( TestingBase ):

    def setUp( self ):
        self.config_for_test()

    def test_make( self ):
        self.skipTest('todo')


if __name__ == '__main__':
    pass


"""
Created by adam on 2/26/20
"""
__author__ = 'adam'

from CanvasHacks.Messaging.skaa import make_prompt_and_response
from CanvasHacks.Models.QuizModels import StoredDataFileMixin
from CanvasHacks.Models.student import Student
from CanvasHacks.Processors.cleaners import TextCleaner
from CanvasHacks.Repositories.interfaces import IContentRepository
from CanvasHacks.Repositories.mixins import SelectableMixin, StudentWorkMixin, FrameStorageMixin
from CanvasHacks.Repositories.submissions import AssignmentSubmissionRepository
from CanvasHacks.Text.stats import WordCount
from CanvasHacks.Widgets.AssignmentSelection import make_selection_button
import pandas as pd

if __name__ == '__main__':
    pass


class AssignmentRepository( IContentRepository, StoredDataFileMixin, StudentWorkMixin, SelectableMixin, FrameStorageMixin ):
    """Manages the data for a non-quiz type unit
    """

    def __init__( self, activity, course=None ):
        self.course = course
        self.activity = activity

        # An assignment will only come in with a 'body' attribute
        # to line this up with things that use questions (i.e., quizzes)
        self.body_column_name = AssignmentSubmissionRepository.body_column_name
        self.question_columns = [ self.body_column_name ]

        # The cleaner class that will be called to
        # remove html and other messy stuff from student
        # work
        self.text_cleaner = TextCleaner()

        self.analyzer = WordCount()


    def process( self, student_work_frame ):
        self.data = student_work_frame
        self._cleanup_data()

    def _cleanup_data( self ):
        """This is abstracted out so it can be
        called independently for use with test data
        """
        prev_len = len( self.data )
        # todo copy dataframe or run on original data?
        self.data = self.data[ self.data.grade != 'complete' ].copy( deep=True )
        # self.data = [j for j in self.data if j[0].grade != 'complete']
        print( "Removed {} rows which have already been graded".format( prev_len - len( self.data ) ) )

        # Remove html and other artifacts from student answers
        # DO NOT UNCOMMENT UNTIL CAN-59 HAS BEEN FULLY TESTED
        for c in self.question_columns:
            self.data[c] = self.data.apply(lambda x: self.text_cleaner.clean(x[c]), axis=1)


        # We set to student id to make look ups easier
        self.data.set_index( 'student_id', inplace=True )

    def get_student_work( self, student_id ):
        try:
            return self.data.loc[ student_id ]
        except (ValueError, KeyError):
            # The student id may not be set as the index, depending
            # on the source of the data
            return self.data.set_index( 'student_id' ).loc[ student_id ]

    def get_formatted_work_by( self, student_id ):
        """Returns all review entries by the student, formatted for
        sending out for review or display"""
        work = self.get_student_work( student_id )
        # narrow down to just the relevant columns
        rs = [ { 'prompt': column_name, 'response': work[ column_name ] } for column_name in
               self.question_columns ]
        r = make_prompt_and_response( rs )
        return self._check_empty( r )

    def make_question_selection_buttons( self ):
        """Given a repository containing a dataframe and a
        list of names in question_names, this will allow to select
        which questions are used for things"""
        buttons = [ ]
        for q in self.question_names:
            b = make_selection_button( q, q, self.get_selections, self.select, self.deselect, '100%' )
            buttons.append( b )

    @property
    def points_per_question( self ):
        return self.assignment.points_possible / len( self.question_columns )

    @property
    def assignment( self ):
        """Returns the canvasapi.unit.Assignment object associated
        with this repository.
        Automatically initializes it if not set
        """
        try:
            if self._assignment:
                pass
        except AttributeError:
            self._assignment = self.course.get_assignment( self.activity.id )
        return self._assignment

    @property
    def submitters( self ):
        """returns a list of student objects for whom work has been submitted"""
        return [ Student( s ) for s in self.student_ids ]

    @property
    def submitter_ids( self ):
        """Returns a list of canvas ids of students who have submitted the unit"""
        # try:
        return list( set( self.data.reset_index().student_id.tolist() ) )
        # except (ValueError, KeyError):

    def word_counts( self ):
        d = []
        for i, row in self.data.iterrows():
            s = {'student_id': i }
            for c in self.question_columns:
                s[c] = self.analyzer.analyze(row[c])
            d.append(s)
        return pd.DataFrame(d)



"""
Created by adam on 12/20/19
"""
__author__ = 'adam'
from canvasapi.user import User
from CanvasHacks.Models.model import Model

from sqlalchemy import Column, Integer, String
from sqlalchemy.ext.declarative import declarative_base

from CanvasHacks import environment as env
from CanvasHacks.Models.model import Model, StoreMixin

from CanvasHacks.DAOs.sqlite_dao import Base


#
# def get_student_id(student_like_thing):
#     """In some contexts poor programming means that there could
#     be either the student id (int), a canvas User, or a Student model"""

def ensure_student(v):
    """
    For functions/methods which require a student object
    but may have received either the object or just the id
    this checks what was passed and returns a student object.
    NB, the object may be new and not stored in db
    :param v:
    :return: Student
    """
    if isinstance(v, Student): return v

    if isinstance(v, User):
        return student_from_canvas_user(v)

    s = Student()
    try:
        # integer and string integer input cases
        s.student_id = int(v)
    except (ValueError, TypeError):
        # dict and object cases
        if not isinstance(v, dict):
            # object case
            v = v.__dict__
        s.handle_kwargs(**v)
    return s


def student_from_canvas_user(user_obj):
    """Makes a student object from a canvas User"""
    s = Student()
    s.student_id = user_obj.id
    # The id will cause a collision with the student.id
    vals = { k : user_obj.attributes[k] for k in user_obj.attributes if k != 'id'}
    s.handle_kwargs(**vals)
    return s


def get_first_name(user):
    """Given a canvas api user object representing a student
    returns the first name
    """
    return user.sortable_name.split(',')[1]


class StoredStudent( Base, StoreMixin ):
    """
    Storable model of student

    as of CAN-55


    """
    __tablename__ = env.STUDENT_TABLE_NAME
    # canvas id
    id = Column( Integer, primary_key=True )
    name = Column( String )
    short_name = Column(String)
    sortable_name = Column(String)
    csun_id = Column(Integer)
    email = Column(String)

    created_at = Column(String)
    integration_id = Column(String)
    login_id = Column(String)

    # def __init__( self, **kwargs ):
    #     self.handle_kwargs(**kwargs)
    #     self.student_id = int(student_id)
    #     # self.name = name
    #
    #     super().__init__( kwargs )

    def __eq__(self, other):
        return self.student_id == other.student_id

    @property
    def student_id( self ):
        """Guaranteed to be an integer since there
        are some cases of creation which may not
        have student_id set as an int
        """
        return int(self.id)

    @property
    def sis_user_id( self ):
        return self.csun_id

    @sis_user_id.setter
    def sis_user_id( self, sid ):
        self.csun_id = sid

    @property
    def canvas_id( self ):
        return self.id

    @property
    def first_name( self ):
        """Returns first name"""
        # if self.short_name is not None:
        # this is always the same as name it seems...
        #     return self.short_name
        if ',' in self.name:
            return self.name.split(',')[1]
        else:
            return self.name.split(' ')[1]



class Student( StoreMixin ):
    """
    NON stored model

    Removed the association with Base
    but did not change the definition of properties
    so won't break anything

    """
    __tablename__ = env.STUDENT_TABLE_NAME
    # Not guaranteed to be an int unless loaded from db
    student_id = Column( Integer, primary_key=True )
    name = Column( String )
    short_name = Column(String)
    sis_user_id = Column(Integer)

    def __init__( self, **kwargs ):
        self.handle_kwargs(**kwargs)
    #     self.student_id = int(student_id)
    #     # self.name = name
    #
    #     super().__init__( kwargs )

    def __eq__(self, other):
        return self.student_id == other.student_id

    @property
    def id( self ):
        """Guaranteed to be an integer since there
        are some cases of creation which may not
        have student_id set as an int
        """
        return int(self.student_id)

    @property
    def first_name( self ):
        """Returns first name"""
        # if self.short_name is not None:
        # this is always the same as name it seems...
        #     return self.short_name
        if ',' in self.name:
            return self.name.split(',')[1]
        else:
            return self.name.split(' ')[1]

if __name__ == '__main__':
    pass



"""
Created by adam on 1/18/20
"""
__author__ = 'adam'

import pandas as pd

class AccessCodeRepo:
    """Manages access codes for assignments which require them"""

    def __init__(self, access_codes_filepath, unit_number):
        print("Loading access codes from ", access_codes_filepath)
        self.data = pd.read_excel(access_codes_filepath)
        self.data = self.data[self.data.unit == unit_number]

    @property
    def review_code(self):
        return self.data[self.data.assignment == 'review'].code.values[0]

    @property
    def metareview_code(self):
        return self.data[self.data.assignment == 'metareview'].code.values[0]



if __name__ == '__main__':
    pass


"""
Created by adam on 3/16/20
"""
__author__ = 'adam'

from CanvasHacks import environment as env
from CanvasHacks.Definitions.base import DiscussionType
from CanvasHacks.Files.FileTools import read_text_block
from CanvasHacks.Models.model import Model
from CanvasHacks.TimeTools import utc_string_to_local_dt, check_is_date

if __name__ == '__main__':
    pass


class Activity( Model ):
    """A wrapper around the canvas provided properties for a quiz which adds
     properties and methods specific to the peer review assignments

    NOT SPECIFIC TO ANY GIVEN STUDENT
    """

    @classmethod
    def is_activity_type( cls, assignment_name ):
        """Given the name of an activity, determines
        whether it is an instance of this unit type"""
        return cls.regex.search( assignment_name.strip().lower() )

    def __init__( self, **kwargs ):
        # when the activity_inviting_to_complete is due
        # ": "2013-01-23T23:59:00-07:00"
        self.due_at = None
        # Set this date if want to give half credit for
        # assignments turned in after this.
        # Normally won't be used, unless manually set
        self.quarter_credit_deadline = None
        # when to lock the activity_inviting_to_complete
        self.lock_at = None
        # // when to unlock the activity_inviting_to_complete
        self.unlock_at = None
        self.points_possible = None
        self._unit_number = None
        self._description_text = ''

        self.creation_dict = { }
        """A dictionary of all the values needed to
        create one of these activities on canvas"""

        super().__init__( **kwargs )

    @property
    def is_quiz_type( self ):
        """Returns whether this is the sort of unit
        that uses a quiz report or not
        """
        try:
            return self.quiz_id > 0
        except AttributeError:
            return False

    @property
    def is_discussion_type( self ):
        """Returns whether this is a discussion.
        Mainly used to simplify decisions elsewhere
        so that we don't have to check for discussion
        type """
        return isinstance( self, DiscussionType )

    @property
    def last_half_credit_date( self ):
        """If a quarter credit deadline has been set
        this will return that value, otherwise it will just
        use the lock date"""
        if self.quarter_credit_deadline is not None:
            return self.quarter_credit_deadline
        return self.lock_at

    @property
    def make_title( self ):
        # if self.unit_number:
        return "Unit {} {}".format( self.unit_number, self.title_base )

    @property
    def assignable_points( self ):
        """How many points are assigned by the reviewer"""
        return self.max_points - self.completion_points

    @property
    def string_due_date( self ):
        """Returns a human readable date for the due date with
        the format yyyy-mm-dd
        """
        if self.due_at is None: return ''
        t = utc_string_to_local_dt( self.due_at )
        return t.date().isoformat()

    @property
    def string_lock_date( self ):
        """Returns a human readable date for the last chance to turn in with
        the format yyyy-mm-dd
        """
        if self.due_at is None: return ''
        t = utc_string_to_local_dt( self.lock_at )
        return t.date().isoformat()

    @staticmethod
    def _check_date( date ):
        """Checks that a value is a pd.Timestamp
        if not, it tries to make it into one"""
        return check_is_date( date )

    @property
    def unit_number( self ):
        try:
            # preference if has been set internally
            return self._unit_number
        except AttributeError:
            return env.CONFIG.unit.unit_number

    @unit_number.setter
    def unit_number( self, unit_number ):
        self._unit_number = unit_number

    @property
    def description( self ):
        try:
            if len( self._description_text ) > 0:
                return self._description_text
            fname = "{}/assignment-templates/{}".format( env.DATA_FOLDER, self.instructions_filename )
            self._description_text = read_text_block( fname )
            return self._description_text

        except (AttributeError, FileNotFoundError) as e:
            print( 'no file for activity', e )
            # In case there is no file for the activity
            return ""

    @description.setter
    def description( self, text ):
        self._description_text = text

    # def creation_dict( self ):
    #     """A dictionary of all the values needed to
    #     create one of these activities on canvas"""
    #     exclude = ['title_base', 'instructions_filename']

    # @property
    # def dates_dict( self ):
    #     return {'assign' self.make_title


"""
Created by adam on 2/18/20
"""
__author__ = 'adam'

if __name__ == '__main__':
    pass

REVIEW_NOTICE_TEMPLATE = """
    Hi {name},
    
    {intro}
    =======================
    
    {responses}
    
    =======================
    {other}
    
     Please make sure you read the instructions in {review_assignment_name} before getting started.
     
     To complete your review, open the quiz named 
             {review_assignment_name} 
             {review_url}
    {access_code_message}
    
    As always, canvas will lie to you about time limits by displaying an ominous, but meaningless in this course, 'Time Elapsed' timer. There is no time-limit other than you must submit your review before 11.59PM on {due_date}. 
    
    You may open and look at the peer review assignment as many times as you like.
    
    Please remember that the person you are reviewing is not the person reviewing you.
        
    Enjoy,
    /a
   
    """

# Used to notify the author that it is time to do the metareview
METAREVIEW_NOTICE_TEMPLATE = """
    Hi {name},
    
    The peer-review responses from the student who read your work are below. Please read them carefully and re-read your original assignment. Then follow the instructions at the bottom of this message to give the reviewer feedback on their review :
    =======================
    
    {responses}
    
    =======================
    {other}
    
    Personally, whenever I get critical feedback on something I've written, my first impression is always wrong ---I either think the problems mentioned are more trivial than they are or go in the opposite direction and overinflate every minor thing into something huge. Thus it may be a good idea to take a break before completing the metareview. 
    
    Please remember that the person who reviewed you is not the person you reviewed.
    
    When you're ready to complete the metareview and give the reviewer feedback, please make sure you read the instructions in {review_assignment_name} before getting started.
     
     To complete your review of their review, open the quiz named 
             {review_assignment_name} 
             {review_url}
     {access_code_message}
    
    As always, canvas will lie to you about time limits by displaying an ominous, but meaningless in this course, 'Time Elapsed' timer. There is no time-limit other than you must submit your review before 11.59PM on {due_date}. 
    
    You may open and look at the metareview assignment as many times as you like.
        
    Enjoy,
    /a
    
    PS, Please let me know if anything seems strange about the review I've sent you. I've tested my software as best I can, but....
   
    """

# Used to send metareview responsese by author to reviewer
METAREVIEW_CONTENT_TEMPLATE = """
 Hi {name},
    
    {intro}
    =======================
    
    {responses}
    
    =======================
    {other}
    
    
    All best wishes,
    /a
   

"""

DISCUSSION_REVIEW_NOTICE_TEMPLATE = """
    Hi {name},
    
    {intro}
    =======================
    
    {responses}
    
    =======================
    {other}
    
     Please make sure you read the instructions in {review_assignment_name} before getting started.
     
     To complete your review, open the assignment named 
             {review_assignment_name} 
             {review_url}
    {access_code_message}
    
    As always, canvas will lie to you about time limits by displaying an ominous, but meaningless in this course, 'Time Elapsed' timer. There is no time-limit other than you must submit your review before 11.59PM on {due_date}. 
    
    You may open and look at the review assignment as many times as you like.
    
    Please remember that the person you are reviewing is not the person reviewing you.
        
    Enjoy,
    /a
   """

# Sending the results of the discussion review to the students
DISCUSSION_REVIEW_FEEDBACK_TEMPLATE = """
Hi {name},
    
    {intro}
    =======================
    
    {responses}
    
    =======================
    {other}
    
    I hope you find this useful. If they have made points with which you disagree, read over your posts and practice trying to see what they had in mind. You might still disagree. But please remember the secondary goal of this is to learn to see your own work through someone else's eyes. That skill is incredibly important, rare, and difficult to learn; it will help you in any industry. So please seize every opportunity for practice. 
    
    All best wishes,
    /a
   
"""


"""
Created by adam on 1/18/20
"""
import canvasapi
import pandas as pd

import CanvasHacks.environment as env
from CanvasHacks.Api.DownloadProcessingTools import extract_body
from CanvasHacks.Repositories.interfaces import ISubmissionRepo

__author__ = 'adam'

from CanvasHacks.Repositories.mixins import ObjectHandlerMixin

class CannotDetermineState(Exception):
    """Raised when can't tell the submission status of a record"""
    pass

class SubmissionRepository( ISubmissionRepo, ObjectHandlerMixin ):
    """Downloads and stores canvasapi.Submission objects
    NB, for many operations on quiz assignments, will need
    to use QuizSubmission objects. That's not this repo
    """

    def __init__( self, assignment: canvasapi.assignment.Assignment ):
        """

        :param assignment: Canvas api unit object for use in downloading
        """
        self.assignment = assignment
        self.download()

    def has_completed( self, student ):
        """
        Determines whether the student has submitted the assignment
        :param student:
        :return: Boolean
        """
        student_id = self._handle_id(student)
        rec = self.get_by_student_id(student_id)
        # state = self.frame.loc['student_id', student_id].workflow_state

        if rec.workflow_state == 'submitted':
            return True

        if rec.workflow_state == 'unsubmitted':
            return False

        raise CannotDetermineState

    def download( self ):
        """Downloads and stores submission objects as a list in self.data"""
        self.data = [ s for s in self.assignment.get_submissions() ]
        for d in self.data:
            d.body = extract_body( d )
        print( "Downloaded {} submissions for unit id {}".format( len( self.data ), self.assignment.id ) )

    @property
    def ungraded_submissions( self ):
        return [ s for s in self.data if s.workflow_state != 'graded' ]

    @property
    def frame( self ):
        """Returns the submissions as a dataframe"""
        s = [ ]
        for r in self.data:
            # Iterating through list of canvasapi.Submission objects
            s.append( { 'course_id': int( r.course_id ),
                        'quiz_id': int( r.assignment_id ),
                        'activity_id': int( r.assignment_id ),
                        'student_id': int( r.user_id ),
                        'user_id': int( r.user_id ),
                        'submission_id': int( r.id ),
                        'attempt': r.attempt,
                        'grade': r.grade,
                        'score': r.score,
                        'workflow_state': r.workflow_state,
                        'late': r.late
                        } )
            # s.append( { 'course_id': int( self.unit.course_id ),
            #             'quiz_id': int( self.unit.id ),
            #             'student_id': int( r[ 'user_id' ] ),
            #             'submission_id': int( r[ 'id' ] ),
            #             'attempt': int( r[ 'attempt' ] )
            #             } )
        return pd.DataFrame( s )

    def get_by_id( self, submission_id: int ):
        """Returns submission object with the id"""
        for s in self.data:
            if s.id == submission_id:
                return s

    def get_by_student_id( self, student_id ):  # , attempt=None ):
        for s in self.data:
            if s.id == student_id:
                # if attempt is None:
                #     return s
                # elif s.attempt == attempt:
                return s


class AssignmentSubmissionRepository( SubmissionRepository ):
    body_column_name = 'body'

    def __init__( self, assignment: canvasapi.assignment.Assignment ):
        """

        :param assignment: Canvas api unit object for use in downloading
        """
        self.assignment = assignment
        self.download()
        self._remove_ignored_users()

    def _remove_ignored_users( self, users=None ):
        """
        Removes all records from data, for each in a list of
        users, either given as a param, or the excluded_users
        list stored on env.CONFIG
        :param users:
        :return:
        """
        users = users if users is not None else env.CONFIG.excluded_users
        prelen = len( self.data )

        # Remove them
        self.data = [ s for s in self.data if s.user_id not in users ]

        removed = prelen - len( self.data )
        if removed > 0:
            print( "Removed {} records belonging to users in the excluded list".format( removed ))

    @property
    def frame( self ):
        """Returns the submissions as a dataframe"""
        s = [ ]
        for r in self.data:
            # Iterating through list of canvasapi.Submission objects
            s.append( { 'course_id': int( r.course_id ),
                        'activity_id': int( r.assignment_id ),
                        'student_id': int( r.user_id ),
                        'user_id': int( r.user_id ),
                        'submission_id': int( r.id ),
                        'attempt': r.attempt,
                        'grade': r.grade,
                        'score': r.score,
                        'workflow_state': r.workflow_state,
                        'late': r.late,
                        self.body_column_name: r.body
                        } )

        return pd.DataFrame( s )


class QuizSubmissionRepository( ISubmissionRepo ):

    def __init__( self, quiz: canvasapi.quiz.Quiz ):
        """
        Should get quiz object with something like
            quiz = environment.CONFIG.course.get_quiz(activity_inviting_to_complete.quiz_id)
        :param quiz: Canvas api unit object for use in downloading
        """
        self.quiz = quiz
        self.download()

    def download( self ):
        """Downloads and stores submission objects as a list in self.data"""
        self.data = [ s for s in self.quiz.get_submissions() ]
        print( "Downloaded {} submissions for quiz id {}".format( len( self.data ), self.quiz.id ) )

    @property
    def frame( self ):
        """Returns the submissions as a dataframe"""
        s = [ ]
        for r in self.data:
            # Iterating through list of canvasapi.Submission objects
            s.append( { 'course_id': int( r.course_id ),
                        'quiz_id': int( r.quiz_id ),
                        'student_id': int( r.user_id ),
                        'user_id': int( r.user_id ),
                        'submission_id': int( r.submission_id ),
                        'attempt': r.attempt,
                        'score': r.score,
                        'workflow_state': r.workflow_state,
                        } )
        return pd.DataFrame( s )

    def get_by_student_id( self, student_id, attempt=None ):
        return [ d for d in self.data if d.user_id == student_id and d.attempt == attempt ][ 0 ]


if __name__ == '__main__':
    pass



"""
Tests that the correct content will get
sent to the reviewer after the metareview is complete
Created by adam on 2/22/20
"""
__author__ = 'adam'

from unittest.mock import MagicMock, patch

import CanvasHacks.environment as env
import CanvasHacks.testglobals
from CanvasHacks.Models.status_record import StatusRecord, InvitationReceivedRecord, FeedbackReceivedRecord
from CanvasHacks.Repositories.status import FeedbackStatusRepository, InvitationStatusRepository
from factories.RepositoryMocks import ContentRepositoryMock
from tests.factories.ModelFactories import student_factory
from tests.factories.PeerReviewedFactories import unit_factory

CanvasHacks.testglobals.use_api = False

from CanvasHacks.DAOs.sqlite_dao import SqliteDAO
from tests.TestingBase import TestingBase

from CanvasHacks.SkaaSteps.SendMetareviewToReviewer import SendMetareviewToReviewer

from faker import Faker

fake = Faker()


# Make sure correct students involved (have more than the 2 in db)
# Make sure correct text is sent
# Make sure addressing params are correct



class TestCallsAllExpected( TestingBase ):
    """Makes sure that everything gets called with expected values.
      Not super diagnostic since many of the mocked calls are
      where we'd actually expect failure. Still, useful for catching
      problems when update code etc"""

    def setUp( self ):
        self.config_for_test()
        self.dao = SqliteDAO()

        self.course = MagicMock()
        self.unit = unit_factory()

    @patch( 'CanvasHacks.SkaaSteps.SendMetareviewToReviewer.RunLogger' )
    @patch( 'CanvasHacks.SkaaSteps.ISkaaSteps.StudentRepository' )
    @patch( 'CanvasHacks.SkaaSteps.SendMetareviewToReviewer.FeedbackFromMetareviewMessenger' )
    @patch( 'CanvasHacks.SkaaSteps.ISkaaSteps.AssociationRepository' )
    @patch( 'CanvasHacks.SkaaSteps.SendMetareviewToReviewer.WorkRepositoryLoaderFactory' )
    def test_run( self, workLoaderMock, assocRepoMock, messengerMock, studentRepoMock, loggerMock ):
        """Check that each student receives the expected message
        containing the correct student's submission
        """
        students = [ student_factory(), student_factory() ]
        submitter_ids = [ s.student_id for s in students ]
        workRepo = MagicMock()
        workRepo.submitter_ids = submitter_ids
        workLoaderMock.make = MagicMock( return_value=workRepo )

        studentRepoMock.download = MagicMock( return_value=students )
        send = True
        obj = SendMetareviewToReviewer( course=self.course, unit=self.unit, is_test=True, send=True )

        # call
        obj.run()

        # check
        workLoaderMock.make.assert_called()
        workLoaderMock.make.assert_called_with( self.unit.metareview, self.course) #, only_new=False, rest_timeout=5 )

        obj.studentRepo.download.assert_called()

        obj.associationRepo.get_by_assessee.assert_called()

        obj.messenger.notify.assert_called()
        obj.messenger.notify.assert_called_with( obj.associations, send )

        loggerMock.log_metareview_feedback_distributed.assert_called()


class TestUnitTests(TestingBase):
    """Unit tests with all dependencies mocked"""
    def setUp( self ):
        self.config_for_test()
        env.CONFIG.semester_name = "T30"
        self.unit = unit_factory()
        self.course = MagicMock()
        self.activity_id = self.unit.review.id
        # self.dao = SqliteDAO()
        self.create_new_and_preexisting_students()
        # Prepare fake work repo to give values to calling  objects
        self.workRepo = ContentRepositoryMock()
        self.workRepo.create_test_content( self.student_ids )
        self.workRepo.add_students_to_data(self.student_ids, make_dataframe=True)

    def test_instantiates_correct_status_repos( self ):
        """The sending of metareview results requires a
        special status repository"""
        obj = SendMetareviewToReviewer( course=self.course, unit=self.unit, is_test=True, send=True )

        self.assertIsInstance(obj.statusRepos, list, "status repos is a list")
        self.assertTrue(len(obj.statusRepos) == 1)
        # self.assertIsInstance(obj.invite_status_repo, InvitationStatusRepository, "Invite repo instantiated")
        self.assertIsInstance( obj.feedback_status_repo, FeedbackStatusRepository, "Correct status repo instantiated" )



class TestFunctionalTests( TestingBase ):
    """Test that works using local data"""

    def setUp( self ):
        self.config_for_test()
        env.CONFIG.semester_name = "T30"
        self.unit = unit_factory()
        self.course = MagicMock()
        self.activity_id = self.unit.review.id
        # self.dao = SqliteDAO()
        self.create_new_and_preexisting_students()
        # Prepare fake work repo to give values to calling  objects
        self.workRepo = ContentRepositoryMock()
        self.workRepo.create_test_content( self.student_ids )
        self.workRepo.add_students_to_data(self.student_ids, make_dataframe=True)

    @patch( 'CanvasHacks.SkaaSteps.SendMetareviewToReviewer.InvitationStatusRepository' )
    @patch( 'CanvasHacks.SkaaSteps.SendMetareviewToReviewer.FeedbackStatusRepository' )
    @patch( 'CanvasHacks.Messaging.base.ConversationMessageSender.send' )
    @patch( 'CanvasHacks.SkaaSteps.ISkaaSteps.StudentRepository' )
    @patch( 'CanvasHacks.SkaaSteps.SendMetareviewToReviewer.WorkRepositoryLoaderFactory' )
    def test_run( self, workLoaderMock, studentRepoMock, messengerMock, feedbackStatusRepoMock, inviteStatusRepoMock ):

        # Prepare fake work repo to give values to calling  objects
        # workRepo = ContentRepositoryMock()
        # workRepo.create_test_content( self.student_ids )
        self.workRepo.submitter_ids = self.student_ids
        self.workRepo.remove_student_records = MagicMock()
        workLoaderMock.make = MagicMock( return_value=self.workRepo )

        # prepare student repo
        students = { s.student_id: s for s in self.students }

        def se( sid ):
            return students.get( sid )

        # set up mocks on instantiated object
        obj = SendMetareviewToReviewer( course=self.course, unit=self.unit, is_test=True, send=True )
        obj.studentRepo.get_student = MagicMock( side_effect=se )
        obj.studentRepo.download = MagicMock( return_value=self.students )

        # Have to do this after object creation so that we can use the
        # same in-memory db
        self.session = obj.dao.session
        # Sets up data for the association repo to use
        self.preexisting_pairings = self.create_preexisting_review_pairings( self.unit.initial_work.id, self.students )

        # call
        obj.run()

        # check
        self.assertEqual( len( obj.associations ), len( self.students ), "Correct number of students notified" )

        # Check that filtered previously notified
        # self.workRepo.remove_student_records.assert_called()

        # ================== Events on Messenger
        # Check that mocked objects were called with expected data
        messengerMock.assert_called()
        self.assertEqual( messengerMock.call_count, len( self.students ),
                          "Send method called expected number of times" )
        messenger_args = [ (c[ 1 ][ 'student_id' ], c[ 1 ][ 'subject' ], c[ 1 ][ 'body' ]) for c in messengerMock.call_args_list ]
        # print(args)

        # Status repo calls on messenger
        for repo in obj.messenger.status_repositories:
            repo.record.assert_called()
            # obj.messenger.status_repositories.record_invited.assert_called()
            call_list = repo.record.call_args_list
            # call_list = obj.messenger.status_repositories.record_invited.call_args_list
            status_args = [ c[ 0 ][ 0 ] for c in call_list ]
            self.assertEqual( len( self.students ), len( call_list ),
                              "Both status repos called expected number of times" )
            for sid in self.student_ids:
                self.assertIn( sid, status_args, "Both status repos called on all students" )

        # student repo calls on messenger
        for sid in self.student_ids:
            obj.messenger.student_repository.get_student.assert_any_call( sid )

        # Check the content sent
        for record in obj.associations:
            # print(record.assessee_id)
            review_text_by_author = self.workRepo.get_formatted_work_by( record.assessee_id )
            # see if sent to assessor
            sent_text = [ t[ 2 ] for t in messenger_args if t[ 0 ] == record.assessor_id ][ 0 ]
            rx = r'{}'.format( review_text_by_author )
            self.assertRegex( sent_text, rx, "Author's review of review in message sent to reviewer" )



    @patch( 'CanvasHacks.Messaging.base.ConversationMessageSender.send' )
    @patch( 'CanvasHacks.SkaaSteps.ISkaaSteps.StudentRepository' )
    @patch( 'CanvasHacks.SkaaSteps.SendMetareviewToReviewer.WorkRepositoryLoaderFactory' )
    def test_status_updated( self, workLoaderMock, studentRepoMock, messengerMock ):
        # Prepare fake work repo to give values to calling  objects
        # workRepo = ContentRepositoryMock()
        # workRepo.create_test_content( self.student_ids )
        self.workRepo.submitter_ids = self.student_ids
        self.workRepo.remove_student_records = MagicMock()
        workLoaderMock.make = MagicMock( return_value=self.workRepo )

        # prepare student repo
        students = { s.student_id: s for s in self.students }

        def se( sid ):
            return students.get( sid )

        # set up mocks on instantiated object
        obj = SendMetareviewToReviewer( course=self.course, unit=self.unit, is_test=True, send=True )
        obj.studentRepo.get_student = MagicMock( side_effect=se )
        obj.studentRepo.download = MagicMock( return_value=self.students )

        # Have to do this after object creation so that we can use the
        # same in-memory db
        self.session = obj.dao.session
        # Sets up data for the association repo to use
        self.preexisting_pairings = self.create_preexisting_review_pairings( self.unit.initial_work.id, self.students )

        r = self.session.query(FeedbackReceivedRecord).filter(FeedbackReceivedRecord.activity_id == self.unit.review.id).all()
        self.assertTrue(len(r) == 0, "No sent feedback on review beforehand")
        r2 = self.session.query(InvitationReceivedRecord).filter(InvitationReceivedRecord.activity_id == self.unit.metareview.id).all()
        self.assertTrue(len(r) == 0, "No sent invitations to metareview beforehand")

        # call
        obj.run()

        # check
        self.assertEqual( len( obj.associations ), len( self.students ), "Correct number of students notified" )

        # check marked as sent
        for status_repo in obj.statusRepos:
            self.assertEqual( len(self.students), len( status_repo.previously_received), "all students have updated status" )

        # Check the content sent
        for record in obj.associations:
            # we need to check each of the reviewers to see that results has a value
            status_record = self.session\
                .query( FeedbackReceivedRecord )\
                .filter( FeedbackReceivedRecord.activity_id == self.unit.review.id )\
                .filter(FeedbackReceivedRecord.student_id == record.assessor_id)\
                .one_or_none()
            self.assertIsNotNone(status_record, "Feedback Record exists")
            self.assertIsNotNone(status_record.sent_at, "Value set for timestamp")

            # invite_status_record = self.session\
            #     .query( InvitationReceivedRecord )\
            #     .filter( InvitationReceivedRecord.activity_id == self.unit.metareview.id )\
            #     .filter(InvitationReceivedRecord.student_id == record.assessor_id)\
            #     .one_or_none()
            # self.assertIsNotNone(invite_status_record, "Invitation Record exists")
            # self.assertIsNotNone(invite_status_record.sent_at, "Value set for timestamp")


    # @patch( 'CanvasHacks.SkaaSteps.ISkaaSteps.FeedbackStatusRepository' )
    @patch( 'CanvasHacks.Messaging.base.ConversationMessageSender.send' )
    @patch( 'CanvasHacks.SkaaSteps.ISkaaSteps.StudentRepository' )
    @patch( 'CanvasHacks.SkaaSteps.SendMetareviewToReviewer.WorkRepositoryLoaderFactory' )
    def test_run_some_already_notified( self, workLoaderMock, studentRepoMock, messengerMock):
        # Set up status repo with some students already notified
        num_previously_sent = fake.random.randint( 1, len( self.students ) - 1 )
        previously_sent = self.student_ids[ : num_previously_sent ]
        # previously_sent = fake.random.choices(self.student_ids, k=num_previously_sent)
        to_notify = [ s for s in self.student_ids if s not in previously_sent ]
        num_to_notify = len( to_notify )
        # feedbackStatusRepoMock.previously_sent_result = MagicMock( return_value=previously_sent )

        # Prepare fake work repo to give values to calling  objects
        # workRepo = ContentRepositoryMock()
        # workRepo.create_test_content( self.student_ids )

        # prepare student repo
        students = { s.student_id: s for s in self.students }

        def se( sid ):
            return students.get( sid )

        # set up mocks on instantiated object
        obj = SendMetareviewToReviewer( course=self.course, unit=self.unit, is_test=True, send=True )
        obj.studentRepo.get_student = MagicMock( side_effect=se )
        obj.studentRepo.download = MagicMock( return_value=self.students )

        # Have to do this after object creation so that we can use the
        # same in-memory db
        self.session = obj.dao.session
        # Sets up data for the association repo to use
        self.preexisting_pairings = self.create_preexisting_review_pairings( self.unit.initial_work.id, self.students )
        # Set up previous notifications

        for sid in previously_sent:
            obj.feedback_status_repo.record(sid)

        authors_with_notified_reviewers = [ r.assessee_id for r in self.preexisting_pairings if r.assessor_id in previously_sent ]
        authors_without_notified_reviewers = [ s for s in self.student_ids if s not in authors_with_notified_reviewers ]

        # have to fake this since using a dummy
        self.workRepo.submitter_ids = authors_without_notified_reviewers
        # workRepo.remove_student_records = MagicMock()
        workLoaderMock.make = MagicMock( return_value=self.workRepo )

        # call
        obj.run()

        # check
        self.assertEqual( num_to_notify, len( obj.associations ), "Correct number of students ({}) notified when {} have already been notified".format(len( obj.associations ), num_to_notify ) )

        # Check that filtered previously notified
        # Have to actually check repo.testText since can't mock
        # self.assertEqual( len( workRepo.testText ), num_to_notify )
        # for sid in previously_sent:
        #     self.assertNotIn( sid, workRepo.testText.keys() )
        # workRepo.remove_student_records.assert_called()
        # workRepo.remove_student_records.assert_called_with(previously_sent)

        # ================== Events on Messenger
        # Check that mocked objects were called with expected data
        messengerMock.assert_called()
        self.assertEqual( messengerMock.call_count, num_to_notify, "Send method called expected number of times" )
        messenger_args = [ (c[ 1 ][ 'student_id' ], c[ 1 ][ 'subject' ], c[ 1 ][ 'body' ]) for c in
                           messengerMock.call_args_list ]
        # print(args)

        # # Status repo calls on messenger
        # obj.messenger.status_repositories.record.assert_called()
        # # obj.messenger.status_repositories.record_invited.assert_called()
        # call_list = obj.messenger.status_repositories.record.call_args_list
        # # call_list = obj.messenger.status_repositories.record_invited.call_args_list
        # status_args = [ c[ 0 ][ 0 ] for c in call_list ]
        # self.assertEqual( num_to_notify, len( call_list ), "Status repo record called expected number of times" )
        # for sid in self.student_ids:
        #     if sid not in previously_sent:
        #         self.assertIn( sid, status_args, "StatusRepo.record called on non prev notified students" )
        #     else:
        #         self.assertNotIn( sid, status_args, "StatusRepo.record not called on prev notified students" )

        # student repo calls on messenger
        # this doesn't work because we record by who received
        # for sid in self.student_ids:
        #     if sid not in previously_sent:
        #         obj.messenger.student_repository.get_student.assert_any_call( sid )

        # Check the content sent
        for record in obj.associations:
            # print(record.assessee_id)
            review_text_by_author = self.workRepo.get_formatted_work_by( record.assessee_id )
            # see if sent to assessor
            sent_text = [ t[ 2 ] for t in messenger_args if t[ 0 ] == record.assessor_id ][ 0 ]
            rx = r'{}'.format( review_text_by_author )
            self.assertRegex( sent_text, rx, "Author's review of review in message sent to reviewer" )

    @patch( 'CanvasHacks.Messaging.base.ConversationMessageSender.send' )
    @patch( 'CanvasHacks.SkaaSteps.ISkaaSteps.StudentRepository' )
    @patch( 'CanvasHacks.SkaaSteps.SendMetareviewToReviewer.WorkRepositoryLoaderFactory' )
    def test_run_some_not_turned_in( self, workLoaderMock, studentRepoMock, messengerMock):
        # Set up status repo with some students already notified
        num_unsubmitted = fake.random.randint( 1, len( self.students ) - 1 )
        submitted = self.student_ids[ : num_unsubmitted ]
        # previously_sent = fake.random.choices(self.student_ids, k=num_previously_sent)
        to_notify = [ s for s in self.student_ids if s in submitted ]
        num_to_notify = len( to_notify )
        # feedbackStatusRepoMock.previously_sent_result = MagicMock( return_value=previously_sent )

        # Prepare fake work repo to give values to calling  objects
        # workRepo = ContentRepositoryMock()
        # workRepo.create_test_content( self.student_ids )
        self.workRepo.submitter_ids = to_notify
        workLoaderMock.make = MagicMock( return_value=self.workRepo )

        # prepare student repo
        students = { s.student_id: s for s in self.students }

        def se( sid ):
            return students.get( sid )

        # set up mocks on instantiated object
        obj = SendMetareviewToReviewer( course=self.course, unit=self.unit, is_test=True, send=True )
        obj.studentRepo.get_student = MagicMock( side_effect=se )
        obj.studentRepo.download = MagicMock( return_value=self.students )

        # Have to do this after object creation so that we can use the
        # same in-memory db
        self.session = obj.dao.session
        # Sets up data for the association repo to use
        self.preexisting_pairings = self.create_preexisting_review_pairings( self.unit.initial_work.id, self.students )

        # call
        obj.run()

        # check
        self.assertEqual( num_to_notify, len( obj.associations ), "Correct number of students ({}) notified when {} have already been notified".format(len( obj.associations ), num_to_notify ) )

        # ================== Events on Messenger
        # Check that mocked objects were called with expected data
        messengerMock.assert_called()
        self.assertEqual( messengerMock.call_count, num_to_notify, "Send method called expected number of times" )
        messenger_args = [ (c[ 1 ][ 'student_id' ], c[ 1 ][ 'subject' ], c[ 1 ][ 'body' ]) for c in messengerMock.call_args_list ]

        # Check the content sent
        for record in obj.associations:
            # get original author's feedback on their review
            review_text_by_author = self.workRepo.get_formatted_work_by( record.assessee_id )
            # see if sent to assessor
            sent_text = [ t[ 2 ] for t in messenger_args if t[ 0 ] == record.assessor_id ][ 0 ]
            rx = r'{}'.format( review_text_by_author )
            self.assertRegex( sent_text, rx, "Author's review of review in message sent to reviewer" )




"""
Created by adam on 3/16/20
"""
__author__ = 'adam'

import re

from CanvasHacks.Definitions.activity import Activity
from CanvasHacks.Definitions.base import DiscussionType
from CanvasHacks.Definitions.groups import DiscussionGroup, ReviewType
from CanvasHacks.GradingCorrections.penalities import NoLatePenalty
from CanvasHacks.GradingMethods.nonempty import CreditForNonEmptyOLD
from CanvasHacks.Models.QuizModels import QuizDataMixin, StoredDataFileMixin

if __name__ == '__main__':
    pass


class DiscussionForum(Activity, DiscussionType,  DiscussionGroup,  StoredDataFileMixin ):
    """Representation of the main discussion forum"""
    title_base = "Main discussion"
    instructions_filename = 'discussion-forum-instructions.txt'
    regex = re.compile( r"\bforum\b" )
    creation_type = 'discussion'

    def __init__( self, **kwargs ):
        super().__init__( **kwargs )

        # The objects which will be used to penalize late assignments
        self.penalizers = [NoLatePenalty()]

        self.corrections = []

        # The object which will be used to penalize late assignments
        # todo deprecated
        # self.penalizer = self.penalizers[0]

        # The object which will be used to assign the score
        self.grade_methods = [CreditForNonEmptyOLD( min_words=2, count_stopwords=True )]
        # self.grade_method = self.grade_method[0]

        try:
            # todo deprecated
            self.penalizer = self.penalizers[ 0 ]
            self.grade_method = self.grade_methods[ 0 ]
        except AttributeError as e:
            print(e)

    def create_on_canvas( self, course ):
        course.create_quiz()

    @property
    def topic_id( self ):
        try:
            return self.discussion_topic[ 'id' ]
        except KeyError:
            print( "No topic id set on discussion " )


class DiscussionReview( Activity, DiscussionType,  ReviewType, DiscussionGroup, QuizDataMixin, StoredDataFileMixin ):
    """Representation of the peer review of the main discussion forum"""
    title_base = "Discussion review"
    instructions_filename = 'discussion-review-instructions.txt'
    regex = re.compile( r"\bdiscussion review\b" )
    creation_type = 'survey'

    def __init__( self, **kwargs ):
        super().__init__( **kwargs )
        self.email_intro = "Here are the discussion forum posts from another student for you to review:"

        self.corrections = [ ]

        # The objects which will be used to penalize late assignments
        self.penalizers = [ NoLatePenalty() ]

        # The object which will be used to assign the score
        # NB, min_words is 1 for now so as to not create problems with multiple-choice answers
        # This could be fixed in CAN-57
        # The object which will be used to assign the score
        self.grade_methods = [ CreditForNonEmptyOLD( min_words=2, count_stopwords=True ) ]

        try:
            # todo deprecated
            self.penalizer = self.penalizers[ 0 ]
            self.grade_method = self.grade_methods[ 0 ]
        except AttributeError as e:
            print(e)

    @property
    def email_subject( self ):
        """Since unit number will be set after initialization
        we need to do this as a property"""
        return "Unit {} peer-review of discussion forum posts".format( self.unit_number )


"""
Created by adam on 3/2/20
"""
__author__ = 'adam'


from CanvasHacks.Errors.review_pairings import NoReviewPairingFound
from CanvasHacks.Messaging.discussions import FeedbackFromDiscussionReviewMessenger
from CanvasHacks.Repositories.factories import WorkRepositoryLoaderFactory
from CanvasHacks.Repositories.status import StatusRepository, FeedbackStatusRepository
from CanvasHacks.SkaaSteps.ISkaaSteps import IStep
from CanvasHacks.Messaging.skaa import MetareviewInvitationMessenger
from CanvasHacks.Logging.run_data import RunLogger
from CanvasHacks.Errors.data_ingestion import NoNewSubmissions

__author__ = 'adam'


class SendDiscussionReviewToPoster(IStep):
    """Handles loading the submitted reviews and routing them to the authors
    """

    def __init__(self, course=None, unit=None, is_test=None, send=True, **kwargs):
        """
        :param course:
        :param unit:
        :param is_test:
        :param send: Whether to actually send the messages
        """
        super().__init__(course, unit, is_test, send, **kwargs)

        # The activity whose results we are going to be doing something with
        self.activity = unit.discussion_review

        # The activity which we are sending feedback about
        self.activity_notifying_about = unit.discussion_review

        # The activity whose id is used to store review pairings for the whole SKAA
        # We use the discussion review so that the association repo records,
        # invite status records and feedback status records will have the same
        # id, which makes reasoning about them easier
        self.activity_for_review_pairings = unit.discussion_review

        # did it this way for unit 2; leaving commented version in case have to go back
        # self.activity_for_review_pairings = unit.discussion_forum

        self.associations = [ ]

        self._initialize()

        self.feedback_status_repo = FeedbackStatusRepository( self.dao, self.activity_notifying_about )
        self.statusRepos = [self.feedback_status_repo]

    def run(self, **kwargs):
        """
        Retrieves submitted reviews and sends them to the authors
        along with instructions for metareview
        Possible kwarg parameters include:
            only_new=False
            rest_timeout=5
        :return:
        """
        try:
            self._load_step( **kwargs)

            self._assign_step()

            self._message_step()

        except NoNewSubmissions:
            # Check if new submitters, bail if not
            print( "No new submissions" )
            RunLogger.log_no_submissions(self.activity)

        except Exception as e:
            print(e)

    def _message_step( self ):
        """
        Handle sending the results of the review to the original author
        :return:
        """
        self.messenger = FeedbackFromDiscussionReviewMessenger( self.unit, self.studentRepo, self.work_repo, self.statusRepos )

        sent = self.messenger.notify( self.associations, self.send )
        self.display_manager.number_sent = sent

        # Log the run
        msg = "Sent {} peer review results \n {}".format( len( self.associations ), self.associations )
        # Note we are distributing the material for the metareview, that's
        # why we're using that activity_inviting_to_complete.
        RunLogger.log_review_feedback_distributed( self.activity_notifying_about, msg )

    def _assign_step( self ):
        """
        Do stuff with the review assignments
        :return:
        """
        self._filter_notified()

        # Filter the review pairs to just those in the work repo.
        for student_id in self.work_repo.submitter_ids:
            try:
                # Work repo contains submitted peer reviews. Thus we look up
                # review pairings where a student submitting the peer review assignment
                # is the assessor
                records = self.associationRepo.get_by_assessor( self.activity_for_review_pairings, student_id )

                if records is None:
                    raise NoReviewPairingFound(student_id)
                else:
                    self.associations.append( records )
            except NoReviewPairingFound:
                print("No review pairing fround for student id {} ".format(student_id))
                # todo Handler for this situation

        self.display_manager.number_to_send = self.associations

    def _load_step( self, **kwargs ):
        """
        Loads applicable data
        kwarg params may include
            download=True/False (default true): Whether to download or load from disk
        :param kwargs:
        :return:
        """
        self.work_repo = WorkRepositoryLoaderFactory.make( self.activity, self.course, **kwargs )
        self.display_manager.initially_loaded = self.work_repo.data

    def _filter_notified( self ):
        """
        The downloaded store of student work contains reviews completed
        by every reviewer who has turned in the assignment.
        We need to remove reviewers who have already had their work sent to
        the original poster.
        That's the job of this method.
        :return:
        """
        records = self.feedback_status_repo.reviewers_with_authors_sent_feedback
        # We can remove those from the repo
        self.work_repo.remove_student_records( records )
        self.display_manager.post_filter = self.work_repo.data





if __name__ == '__main__':
    pass


"""
Created by adam on 2/25/20
"""
from CanvasHacks.Models.model import StoreMixin

__author__ = 'adam'

if __name__ == '__main__':
    pass


class MessageDataCreationError(Exception, StoreMixin):
    """"Something went wrong in creating the data that would be
    used to format the message
    """
    def __init__( self, review_assignment, **kwargs ):
        self.review_assignment = review_assignment
        self.handle_kwargs(**kwargs)


class MessageSendError(Exception):
    """Wraps an error response from server when an
    attempt to send a message fails"""
    pass

class NoStudentsNeedNotification(Exception):
    """
    Raised when no students need to be notified
    for the current step
    """
    pass


"""
Created by adam on 2/22/20
"""
__author__ = 'adam'

import pandas as pd
from faker import Faker

from CanvasHacks.Models.model import StoreMixin
from CanvasHacks.Repositories.interfaces import IContentRepository
from CanvasHacks.Repositories.students import StudentRepository
from factories.PeerReviewedFactories import content_repo_data_object_factory

fake = Faker()

if __name__ == '__main__':
    pass


class StudentRepositoryMock( StudentRepository, StoreMixin ):

    def __init__( self, **kwargs ):
        self.handle_kwargs( **kwargs )

    def get_student( self, canvas_id ):
        return self.student

    def get_student_name( self, canvas_id ):
        return self.student_name


class ContentRepositoryMock( IContentRepository, StoreMixin ):
    """Mainly used to avoid creating whole dataframes
    for testing
    """

    def __init__( self, **kwargs ):
        self.handle_kwargs( **kwargs )
        self.data = [ ]

    def get_formatted_work_by( self, student_id ):
        return self.testText.get( student_id )
        # return [a[1] for a in self.testText if a[0] == student_id][0]

    def create_test_content( self, student_ids ):
        """Creates a dictionary with a content / body entry as the value
        for each id passed in and stores it on testText"""
        self.testText = { sid: fake.paragraph() for sid in student_ids }

    # @property
    # def submitter_ids( self ):
    #     return [k for k in self.testText.keys()]

    def remove_student_records( self, student_ids ):
        self.testText = { k: self.testText[ k ] for k in self.testText.keys() if k not in student_ids }

    def add_students_to_data( self, student_ids, make_dataframe=False, **kwargs ):
        """Populates self.data with records of the expected type"""
        for s in student_ids:
            d = content_repo_data_object_factory( student_id=s, **kwargs )
            d[ 'body' ] = self.testText.get( s )
            self.data.append( d )
        if make_dataframe:
            self.data = pd.DataFrame( self.data )

    def create_quiz_repo_data( self, student_ids, submitted_at, make_dataframe=False, num_question_columns=5, **kwargs ):
        """Simulates being a content repo holding quiz data
        """
        self.question_columns = [ (fake.random.randint(1, 99999999), fake.word()) for _ in range( 0, num_question_columns ) ]
        fixed_fields = [ 'course_id', 'quiz_id']
        variable_fields = ['attempt', 'submission_id']

        course_id = kwargs['course_id'] if 'course_id' in kwargs else fake.random.randint(1, 999999)

        quiz_id = kwargs[ 'quiz_id' ] if 'quiz_id' in kwargs else fake.random.randint( 1, 999999 )

        for sid in student_ids:
            d = {
                'student_id': sid,
                'submitted_at': submitted_at,
                # some things look for this still
                'submitted': submitted_at,
                'course_id': course_id,
                'quiz_id': quiz_id
            }
            for f in variable_fields:
                if f in kwargs.keys():
                    d[f] = kwargs[f]
                else:
                    d[f] = fake.random.randint(11111, 9999999)

            for qid, column_name in self.question_columns:
                d[  column_name ] = fake.paragraph()

            self.data.append( d )

        if make_dataframe:
            self.data = pd.DataFrame(self.data)



"""
Various utilities for project

Created by adam on 1/19/20
"""
from functools import wraps

import pandas as pd

__author__ = 'adam'





if __name__ == '__main__':
    pass


def ensure_timestamps(f):
    """Makes sure that everything passed into the function
    is either a pd.Timestamp or pd.Timedelta
    """
    @wraps(f)
    def decorated(*args, **kwargs):
        args = list(args)
        i=0
        for a in args:
            if not isinstance(a, pd.Timestamp) and not isinstance(a, pd.Timedelta):
                args[i] = pd.to_datetime(a)
            i+=1

        for k in kwargs.keys():
            if not isinstance(kwargs[k], pd.Timestamp) and not isinstance(a, pd.Timedelta):
                kwargs[k] = pd.to_datetime(kwargs[k])
        return f(*args, **kwargs)

    return decorated


"""
Created by adam on 10/18/18
"""
__author__ = 'adam'

import seaborn as sns
import wordcloud
from matplotlib import pyplot as plt

sns.set( style="whitegrid" )


def clearplot_function( ):
    """Clears the matplotlib plot to free system resources. Call in next cell after plot"""
    try:
        plt.close( plt.gcf( ) ); plt.close( 'all' )
    except:
        pass


def rotate_x_labels(axis, degrees=45):
    for tick in axis.get_xticklabels():
        tick.set_rotation(degrees)

def draw_cloud( wordlist, title, font_size=20 ):
    """Given a list of words, it plots a wordcloud"""
    # make into block of text
    words = ' '.join( wordlist )

    wc = wordcloud.WordCloud(
        width=2500,
        height=2000
    ).generate( words )
    plt.figure( 1, figsize=(13, 13) )
    plt.imshow( wc )
    plt.axis( 'off' )
    plt.title( title )
    ax = plt.gca()
    ax.title.set_fontsize( font_size )
    plt.show()


def draw_cloud_from_freqs( freqDistObj, title, font_size=20 ):
    """Given the frequency object with keys like w1, w2,
    this plots a wordcloud from those frequencies
    """
    # f = { }
    # w = "w%s" % week_num
    # for k in freqs[ w ].freqDist:
    #     f[ k ] = freqs[ w ].freqDist.freq( k )

    wc = wordcloud.WordCloud( width=2500, height=2000 ).generate_from_frequencies( freqDistObj )
    plt.figure( 1, figsize=(13, 13) )
    plt.imshow( wc )
    plt.axis( 'off' )
    plt.title( title )
    ax = plt.gca()
    ax.title.set_fontsize( font_size )
    plt.show()


def draw_cumulative_freq( freqDistObj, title, max_terms=30, font_size=20 ):
    """Plots cumulative frequencies of terms"""
    # week = "w%s" % week_num

    fig = plt.gcf()
    ax = plt.gca()
    ax.set_title( title )

    #     # rotate the x axis labels to 45 degrees
    #     for tick in ax.get_xticklabels():
    #         tick.set_rotation(45)

    fig.set_figwidth( 13 )
    fig.set_figheight( 5 )
    # change the font size
    for item in ([ ax.title, ax.xaxis.label, ax.yaxis.label ] + ax.get_xticklabels() + ax.get_yticklabels()):
        item.set_fontsize( font_size )

    plt.xticks( rotation=45 )
    freqDistObj.plot( max_terms )

    # freqs[ week ].plot( max_terms )
    fig.tight_layout()


if __name__ == '__main__':
    pass



"""
Created by adam on 12/24/19
"""
import pandas as pd
from unittest import TestCase
from tests.factories.PeerReviewedFactories import test_data_factory, unit_factory
from CanvasHacks.Models.model import Model
from CanvasHacks.Definitions.skaa import InitialWork, Review, MetaReview
from CanvasHacks.Definitions.journal import Journal
from CanvasHacks.Definitions.activity import Activity
from CanvasHacks.Definitions.unit import Unit

__author__ = 'adam'

if __name__ == '__main__':
    pass


class DummyAssignment( Model ):
    def __init__( self, **kwargs ):
        self.attributes = { }
        super().__init__( **kwargs )

        for key, value in kwargs.items():
            self.attributes[ key ] = value


class TestActivity( TestCase ):
    # def test_assignable_points( self ):
    #     self.fail()

    def test_check_date( self ):
        # is date case
        self.assertIsInstance( Activity._check_date( pd.to_datetime( '2019-01-06' ) ), pd.Timestamp )
        # string case
        self.assertIsInstance( Activity._check_date( '2019-01-06' ), pd.Timestamp )

    def test_instantiates_properly( self ):
        test_data = test_data_factory()
        obj = Activity( **test_data[ 'initial' ] )
        self.assertIsInstance( obj.due_at, pd.Timestamp, "due date is timestamp" )
        self.assertIsInstance( obj.open_at, pd.Timestamp, "open date is timestamp" )
        pts = test_data[ 'initial' ][ 'max_points' ] - test_data[ 'initial' ][ 'completion_points' ]
        self.assertEqual( obj.assignable_points, pts, 'assignable points correctly calculated' )


class TestUnit( TestCase ):
    def setUp( self ):
        self.obj = Unit( { }, 1 )

    def test_find_for_unit( self ):
        good = [ DummyAssignment( name="tacos are unit 1 best" ), DummyAssignment( name="Unit 1" ),
                 DummyAssignment( name='tacos are UNIT 1 best' ) ]
        bad = [ DummyAssignment( name="tacos are best" ), DummyAssignment( name="Unit 2" ),
                DummyAssignment( name="unit 10" ), DummyAssignment( name="tacos are UNIT" ) ]
        tests = good + bad
        results = self.obj.find_for_unit( 1, tests )
        self.assertTrue( len( results ) == len( good ) )
        # for r in results:
        #     self.assertIn( r, good )

    def test_find_components( self ):
        # assigns = []
        # vs = ['Unit 1 Content unit','Unit 1 metareview''Unit 1 review']
        # for v in vs:
        #     a = canvasapi.unit.Assignment()
        #     a.name = v
        #     assigns.append(a)
        assigns = [ DummyAssignment( name='Unit 1 Essay', id=5, due_at= pd.to_datetime( '2019-01-06' ) ),
                    DummyAssignment( name='Unit 1 metareview', id=4, due_at= pd.to_datetime( '2019-01-06' ) ),
                    DummyAssignment( name='Unit 1 peer review', id=3, due_at= pd.to_datetime( '2019-01-06' ) )
                    ]
        self.obj.find_components( assigns )
        self.assertEqual( len( self.obj.components ), 3 )

    def test__set_access_code_for_next( self ):
        unit = unit_factory()
        for c in unit.components:
            unit._set_access_code_for_next(c)

        # check
        self.assertEqual(unit.initial_work.access_code_for_next, unit.review.access_code)
        self.assertEqual(unit.review.access_code_for_next, unit.metareview.access_code)


class TestInitialWork( TestCase ):
    def setUp( self ):
        self.test_data = test_data_factory()
        # Create the unit
        # self.initial = InitialWork( **self.test_data[ 'initial' ] )
        # self.review = Review( **self.test_data[ 'review' ] )
        # self.meta = MetaReview( **self.test_data[ 'metareview' ] )
        # self.obj = Assignment( self.initial, self.review, self.meta )

    def test_is_activity_type( self ):
        # Create the unit
        self.assertFalse( InitialWork.is_activity_type( 'Unit 24 metareview' ) )
        self.assertTrue( InitialWork.is_activity_type( 'Unit 24 essay' ) )


class TestReview( TestCase ):
    def setUp( self ):
        self.test_data = test_data_factory()
        # Create the unit
        # self.initial = InitialWork( **self.test_data[ 'initial' ] )
        # self.review = Review( **self.test_data[ 'review' ] )
        # self.meta = MetaReview( **self.test_data[ 'metareview' ] )
        # self.obj = Assignment( self.initial, self.review, self.meta )

    def test_is_activity_type( self ):
        # Create the unit
        self.assertFalse( Review.is_activity_type( 'Unit 24 metareview' ) )
        self.assertTrue( Review.is_activity_type( 'Unit 24 peer review' ) )


class TestMetaReview( TestCase ):
    def setUp( self ):
        self.test_data = test_data_factory()
        # Create the unit
        # self.initial = InitialWork( **self.test_data[ 'initial' ] )
        # self.review = Review( **self.test_data[ 'review' ] )
        # self.meta = MetaReview( **self.test_data[ 'metareview' ] )
        # self.obj = Assignment( self.initial, self.review, self.meta )

    def test_is_activity_type( self ):
        # Create the unit
        self.assertTrue( MetaReview.is_activity_type( 'Unit 24 metareview' ) )
        self.assertFalse( MetaReview.is_activity_type( 'Unit 24 peer review' ) )

# class TestAssignment( TestCase ):
#     def setUp( self ):
#         self.test_data = test_data_factory()
#         # Create the unit
#         self.initial = InitialWork( **self.test_data[ 'initial' ] )
#         self.review = Review( **self.test_data[ 'review' ] )
#         self.meta = MetaReview( **self.test_data[ 'metareview' ] )
#         self.obj = Assignment( self.initial, self.review, self.meta )
#
#     def test_setup( self ):
#         # Create the unit
#         initial = InitialWork( **self.test_data[ 'initial' ] )
#         review = Review( **self.test_data[ 'review' ] )
#         meta = MetaReview( **self.test_data[ 'metareview' ] )
#         obj = Assignment( initial, review, meta )
#
#         self.assertIsInstance( obj.review, Review )

class TestJournal( TestCase ):
    def setUp( self ):
        pass
        # self.test_data = test_data_factory()
        # Create the unit
        # self.initial = InitialWork( **self.test_data[ 'initial' ] )
        # self.review = Review( **self.test_data[ 'review' ] )
        # self.meta = MetaReview( **self.test_data[ 'metareview' ] )
        # self.obj = Assignment( self.initial, self.review, self.meta )

    def test_is_activity_type( self ):
        # Create the unit
        self.assertTrue( Journal.is_activity_type( 'Journal (week 4)' ) )
        self.assertFalse( Journal.is_activity_type( 'Unit 24 peer review' ) )




"""
Determines whether a score needs to be corrected
for various reasons
Created by adam on 3/9/20
"""
__author__ = 'adam'

from CanvasHacks.Errors.grading import StudentUnableToComplete
from CanvasHacks.Errors.review_pairings import NoReviewPairingFound
from CanvasHacks.GradingCorrections.base import IGradeCorrection
from CanvasHacks.Repositories.mixins import ObjectHandlerMixin

if __name__ == '__main__':
    pass


class ForfeitedPointsCorrector(IGradeCorrection):
    """
    Boosts student's score by the points their non turning in
    reviewer forfeited
    """

    def analyze( self, *args, **kwargs ):
        """

        :param args:
        :param kwargs:
        :return:
        """
        pass


class NonSubmittingStudentCorrector(ObjectHandlerMixin):
    """
    For assignments where a student's ability to complete it depends
    on other students doing something, this determines whether a correction
    is needed, e.g., if we're on a metareview but the reviewer never turned it in
    """

    def __init__(self, submissions_repo, pairing_repo):
        """
        :param dependency_work_repo: The assignment a student's ability to complete depends on
        :param pairing_repo: Review pairings that create the dependency
        """
        self.submissions_repo = submissions_repo
        self.pairing_repo = pairing_repo

        # The field in the repo that we will check to determine
        # whether the student depended on did the work
        self.completion_field = 'workflow_state'
        # The value the completion field will have if completed
        self.complete_value = 'submitted'
        self.incomplete_value = 'unsubmitted'


    # def is_able_to_complete( self, student ):
    #     """
    #     Looks up whether the provided student is/was able to complete
    #     the assignment.
    #
    #     Will raise exceptions to make it easier to bubble
    #     them up to various handlers (e.g., one which notifies the blocked
    #     and blocking student)
    #
    #     :param student: Student object or int
    #     :return: boolean
    #     :raises StudentUnableToComplete, NoReviewPairingFound
    #     """
    #     student_id = self._handle_id(student)
    #
    #     ra = self.pairing_repo.get_by_author(student_id)
    #     depends_on = ra.assessor_id
    #
    #     if depends_on is None:
    #         # todo If this is true the student was never assigned a reviewer, what should be done?
    #         raise NoReviewPairingFound(student_id)
    #         # return False
    #
    #     dependency_student_row = self.dependency_work_repo.get_student_work(student_id)
    #
    #     if dependency_student_row[self.completion_field] == self.complete_value:
    #         return True
    #
    #     raise StudentUnableToComplete(blocked_student=student_id, blocking_student=depends_on)
    #



"""
Created by adam on 5/6/19
"""
from unittest import TestCase

from CanvasHacks.GradingCorrections.penalities import get_penalty, HalfLate, QuarterLate, NoLatePenalty
import pandas as pd

__author__ = 'adam'

from TestingBase import TestingBase


class TestNoLatePenalty(TestingBase):
    def setUp(self):
        self.config_for_test()
        self.due_date = pd.to_datetime('2019-02-23 07:59:00')
        self.grace_period = None
        self.obj = NoLatePenalty(self.due_date, self.grace_period)

    def test_get_penalty( self ):
        result = self.obj.get_penalty(self.fake.date_time_this_century())
        self.assertEqual(0, result, "get penalty returns 0")

    def test_get_penalized_score( self ):
        original_score = self.fake.random.randint(1, 111111)
        result = self.obj.get_penalized_score( self.fake.date_time_this_century(), original_score )
        self.assertEqual( original_score, result, "get penalized score returns original score" )

    def test_get_fudge_points( self ):
        original_score = self.fake.random.randint( 1, 111111 )
        result = self.obj.get_fudge_points( self.fake.date_time_this_century(), original_score, {} )
        self.assertEqual( 0, result, "get fudge points returns 0" )


class TestHalfLate(TestingBase):

    def setUp(self):
        self.config_for_test()

        self.due_date = pd.to_datetime('2019-02-23 07:59:00')
        self.grace_period = None
        self.obj = HalfLate(self.due_date, self.grace_period)

    def test_get_penalty( self ):
        test_cases = [
            {
                # full credit case
                'submitted': '2019-02-22 07:59:00',
                # 'due': '2019-02-23 07:59:00',
                'half': '2019-03-01 07:59:00',
                'expect': 0
            },
            {
                # half credit case
                'submitted': '2019-02-24 07:59:00',
                # 'due': '2019-02-23 07:59:00',
                'half': '2019-03-01 07:59:00',
                'expect': .5
            }

        ]

        for t in test_cases:
            self.assertEquals( self.obj.get_penalty( t[ 'submitted' ]), t[ 'expect' ], "Returns expected penalty" )

    def test_get_penalized_score( self ):
        original_score = self.fake.random.randint(1, 2402)
        test_cases = [
            {
                # full credit case
                'submitted': '2019-02-22 07:59:00',
                # 'due': '2019-02-23 07:59:00',
                'half': '2019-03-01 07:59:00',
                'expect': original_score
            },
            {
                # half credit case
                'submitted': '2019-02-24 07:59:00',
                # 'due': '2019-02-23 07:59:00',
                'half': '2019-03-01 07:59:00',
                'expect': original_score * .5
            }
        ]

        for t in test_cases:
            self.assertEquals(t[ 'expect' ], self.obj.get_penalized_score( t[ 'submitted' ], original_score),  "Returns expected score" )

    def test_get_fudge_points_no_penalty( self ):
        submitted_date = self.due_date - pd.Timedelta('{} days'.format(self.fake.random.randint(1, 200)))
        original_score = self.fake.random.randint(1, 2222)
        row = {'submitted' : 'taco'}

        # call
        result = self.obj.get_fudge_points(submitted_date, original_score, row)

        # check
        self.assertEqual(result, 0, "zero fudge points returned")
        self.assertNotIn( row, self.obj.penalized_records, "Row not added to penalized rows" )

    def test_get_fudge_points_penalty( self ):
        submitted_date = self.due_date + pd.to_timedelta( '{} days'.format( self.fake.random.randint( 1, 200 ) ) )
        # not randomizing so don't need to deal with float in checking record
        original_score = 100
        row = { 'submitted': 'taco' }

        # call
        result = self.obj.get_fudge_points( submitted_date, original_score, row )

        # check
        self.assertEqual(result, -50, "fudge points of 50% original returned" )
        expected_rec = {'record': row, 'penalty': 0.5, 'fudge_points': -50}
        self.assertIn( expected_rec, self.obj.penalized_records, "Row added to penalized rows" )


class TestQuarterLate(TestingBase):

    def setUp(self):
        self.config_for_test()

        self.due_date = pd.to_datetime('2019-02-23 07:59:00')
        self.last_half_date = pd.to_datetime('2019-03-01 07:59:00')
        self.grace_period = None
        self.obj = QuarterLate(self.due_date, last_half_credit_date=self.last_half_date, grace_period=self.grace_period)

    def test_get_penalty( self ):
        test_cases = [
            {
                # full credit case
                'submitted': '2019-02-22 07:59:00',
                # 'due': '2019-02-23 07:59:00',
                # 'half': '2019-03-01 07:59:00',
                'expect': 0
            },
            {
                # half credit case
                'submitted': '2019-02-24 07:59:00',
                # 'due': '2019-02-23 07:59:00',
                # 'half': '2019-03-01 07:59:00',
                'expect': .5
            },
            {
                # quarter credit case
                'submitted': '2019-03-02 07:59:00',
                # 'due': '2019-02-23 07:59:00',
                # 'half': '2019-03-01 07:59:00',
                'expect': .25
            },
        ]

        for t in test_cases:
            self.assertEquals( self.obj.get_penalty( t[ 'submitted' ]), t[ 'expect' ], "Returns expected penalty" )

    def test_get_fudge_points_no_penalty( self ):
        submitted_date = self.due_date - pd.Timedelta( '{} days'.format( self.fake.random.randint( 1, 200 ) ) )
        original_score = self.fake.random.randint( 1, 2222 )
        row = { 'submitted': 'taco' }

        # call
        result = self.obj.get_fudge_points( submitted_date, original_score, row )

        # check
        self.assertEqual( result, 0, "zero fudge points returned" )
        self.assertEqual(0, len(self.obj.penalized_records), "Row not added to penalized records")
        # self.assertNotIn( row, self.obj.penalized_records, "Row not added to penalized rows" )

    def test_get_fudge_points_quarter_penalty( self ):
        submitted_date = self.last_half_date + pd.to_timedelta( '1 days')
        original_score = 100
        row = { 'submitted': 'taco' }

        # call
        result = self.obj.get_fudge_points( submitted_date, original_score, row )

        # check
        self.assertEqual( result, -25, "fudge points of 25% original returned" )
        expected_rec = {'record': row, 'penalty': 0.25, 'fudge_points': -25}
        self.assertIn( expected_rec, self.obj.penalized_records, "Record added to penalized rows" )

    def test_get_fudge_points_half_penalty( self ):
        submitted_date = self.due_date + pd.to_timedelta( '1 days' )
        original_score = 100
        row = { 'submitted': 'taco' }

        # call
        result = self.obj.get_fudge_points( submitted_date, original_score, row )

        # check
        self.assertEqual( result, -50, "fudge points of 50% original returned" )
        expected_rec = { 'record': row, 'penalty': 0.5, 'fudge_points': -50 }
        self.assertIn( expected_rec, self.obj.penalized_records, "Record added to penalized rows" )

    # def test_get_penalized_score( self ):
    #     original_score = self.fake.random.randint(1, 2402)
    #     test_cases = [
    #         {
    #             # full credit case
    #             'submitted': '2019-02-22 07:59:00',
    #             # 'due': '2019-02-23 07:59:00',
    #             'half': '2019-03-01 07:59:00',
    #             'expect': original_score
    #         },
    #         {
    #             # half credit case
    #             'submitted': '2019-02-24 07:59:00',
    #             # 'due': '2019-02-23 07:59:00',
    #             'half': '2019-03-01 07:59:00',
    #             'expect': original_score * .5
    #         }
    #     ]

# ------------------------- older
class TestGet_penalty( TestCase ):
    def test_get_penalty( self ):
        test_cases = [
            {
                # full credit case
                'submitted': '2019-02-22 07:59:00',
                'due': '2019-02-23 07:59:00',
                'half': '2019-03-01 07:59:00',
                'expect': 0
            },
            {
                # half credit case
                'submitted': '2019-02-24 07:59:00',
                'due': '2019-02-23 07:59:00',
                'half': '2019-03-01 07:59:00',
                'expect': .5
            },
            {
                # quarter credit case
                'submitted': '2019-03-02 07:59:00',
                'due': '2019-02-23 07:59:00',
                'half': '2019-03-01 07:59:00',
                'expect': .25
            },
        ]

        for t in test_cases:
            self.assertEquals( get_penalty( t[ 'submitted' ], t[ 'due' ], t[ 'half' ] ), t[ 'expect' ] )



"""
Created by adam on 12/20/19
"""
from CanvasHacks.Definitions.skaa import InitialWork, Review, MetaReview
from CanvasHacks.Definitions.unit import Unit

__author__ = 'adam'

from CanvasHacks.Models.student import Student


def determine_overall_grade():
    """Just submitting: 1
        points assigned by reviewer: 1 , 2, 3
        points for just doing review: 1
        points for quality of review (assigned by submitter): 1, 2
        points for reviewing reviewer: 1
    """
    submission = {
        # Points get for just turning in something non-empty
        'automatic': 1
    }

    reviewing = {
        'automatic': 1
    }

    review_review = {
        'automatic': 1
    }


# -------------------------------- Elements of the unit


# ---------------------------------- Student work


class UnitScores( object ):
    """Holds all scores for one student"""

    def __init__( self, assignment: Unit, student: Student ):
        self.student = student
        self.assignment = assignment

        self.initial_complete = 0
        self.initial_assigned = 0

        self.review_complete = 0
        self.review_assigned = 0

        self.meta_complete = 0

    @property
    def total( self ):
        return sum( [
                self.initial_complete,
                self.initial_assigned,
                self.review_complete,
                self.review_assigned,
                self.meta_complete
            ] )


class UnitGradeCalculator( object ):
    """Calculates the unit grades for all students.
 """

    def __init__( self, unit: Unit, submissions ):
        self.assignment = unit
        self.submissions = submissions
        self.scores = []

    def add_submission( self, submission ):
        self.submissions.append( submission )

    def _get_submissions( self, submission_type, submission_list=None ):
        """Returns a list of all submissions of the provided type
        Will search all stored submissions unless pass in a particular list
        (which would normally be the submissions for a particular student)"""
        submission_list = submission_list if submission_list else self.submissions
        return list( filter( lambda x: isinstance( x.activity, submission_type ), submission_list ) )

    @property
    def intial_work_submissions( self ):
        return self._get_submissions( InitialWork )

    @property
    def review_submissions( self ):
        return self._get_submissions( Review )

    @property
    def meta_review_submissions( self ):
        return self._get_submissions( MetaReview )

    @property
    def students( self ):
        """Returns list of unique student objects.
        This is drawn from initial work since students only
        do reviews after they submit something
        todo: make unique ( using list(set(xxx)) stopped working when defined __eq__ on student
        """
        return [ s.submitter for s in self.intial_work_submissions ]

    def get_student_work( self, student: Student ):
        """Returns list containing submissions where the
        student is either the submtter or the reviewer
        """
        return self.get_work_as_submitter( student ) + self.get_work_as_reviewer( student )
        # def check(user, submission):
        #     if submission.submitter.student_id == user.student_id:
        #         return True
        #     if submission.reviewer is not None:
        #         return submission.reviewer.student_id == user.student_id
        #     return False
        # return list( filter( lambda x: check(student, x), self.submissions ))

    def get_work_as_submitter( self, student: Student ):
        """Returns all submissions where the student is the submitter"""
        return list( filter( lambda x: x.submitter == student, self.submissions ) )

    def get_work_as_reviewer( self, student: Student ):
        """Returns all submissions where the student is the reviewer"""

        def check( user, submission ):
            if submission.reviewer is not None:
                return submission.reviewer == user
            return False

        return list( filter( lambda x: check( student, x ), self.submissions ) )

    # def calc_orig_scores( self, student ):
    #     """Calculates the initial work portion of the
    #     score
    #     Includes:
    #         - Credit for submitting the unit
    #         - Credit assigned by the reviewer
    #     """
    #
    # def calc_review_scores( self, student ):
    #     """Calculates the review portion of the score.
    #     This includes:
    #         - credit for completing the review
    #         - credit assigned in the metareview
    #     """
    #     # Get the reviews for the submitter
    #     work = self.get_student_work( student )
    #     reviews = self._get_submissions( Review, work )
    #     for r in reviews:
    #         if r.submitter == student:
    #             # get points assigned by the reviewer
    #             scores[ 'review_assigned' ] = r.submitter_points
    #         elif r.reviewer == student:
    #             # Points for completing the review
    #             scores[ 'review_complete' ] = r.reviewer_points
    #
    # def calc_meta_scores( self, student ):
    #     """Calculates the meta review portion of the score
    #     This includes:
    #         - credit for completing the metareview
    #     """

    def calculate( self ):
        # Get a list of students and
        # populate our score data store
        for student in self.students:
            scores = UnitScores( self.assignment, student )

            work = self.get_student_work( student )

            # Get the points for initial submissions
            initial = self._get_submissions( InitialWork, work )
            # todo error handling if multiple initial subs
            #   self.scores[s.submitter.student_id]
            scores.initial_complete += initial[0].submitter_points

            reviews = self._get_submissions( Review, work )
            for r in reviews:
                if r.submitter == student:
                    # get points assigned by the reviewer
                    scores.review_assigned = r.submitter_points
                elif r.reviewer == student:
                    # Points for completing the review
                    scores.review_complete = r.reviewer_points

            # Metareviews
            meta = self._get_submissions( MetaReview, work )
            for r in meta:
                if r.submitter == student:
                    # todo: is this right?
                    # Points for completing the review
                    scores.meta_complete = r.reviewer_points

                elif r.reviewer == student:
                    # get points assigned to student as reviewer
                    scores.meta_assigned = r.submitter_points

            # Calculate total
            self.scores.append(scores)

            # self.scores[ student.student_id ] = scores
        # # Get the automatic credit/no-credit score
        # for sub in self.submissions:
        #     self.score += sub.points
        # return self.score

        # Now get the points assigned to them by the reviewer
        # self.review_submissions


# ---------------- queries and mechanics (move elsewhere)

if __name__ == '__main__':
    pass



# """
# Created by adam on 12/20/19
# """
# from unittest import TestCase
#
# from faker import Faker
#
# fake = Faker()
#
# __author__ = 'adam'
#
# from CanvasHacks.PeerReviewed.PeerReviewTools import *
#
# from tests.factories.PeerReviewedFactories import *
# from tests.factories.ModelFactories import *
#
#
# class TestAssignmentGradeCalculator( TestCase ):
#     def setUp( self ):
#         # Create the unit
#         self.assignment = unit_factory()
#
#         self.student1 = student_factory()
#         self.student2 = student_factory()
#
#         self.submissions = submissions_factory( self.student1, self.student2, self.assignment )
#
#         self.obj = UnitGradeCalculator( self.assignment, self.submissions )
#
#     def test_get_submissions_for_all_subs( self ):
#         c = self.obj._get_submissions( InitialWork )
#         self.assertIsInstance( c[ 0 ], Submission )
#
#     def test_get_work_as_submitter( self ):
#         # check starting point
#         self.assertEqual( 6, len( self.obj.submissions ), "Starting with correct number" )
#
#         work = self.obj.get_work_as_submitter( self.student1 )
#         self.assertEqual( 3, len( work ), "Returns expected number of subs" )
#
#         for w in work:
#             self.assertIsInstance( w, Submission, "Returns expected type" )
#             # if isinstance(w, InitialWork):
#             self.assertEqual( w.submitter.student_id, self.student1.student_id,
#                               "Student is submitter of the work (including cases where another student reviewed it)" )
#
#     def test_get_work_as_reviewer( self ):
#         # check starting point
#         self.assertEqual( 6, len( self.obj.submissions ), "Starting with correct number" )
#
#         work = self.obj.get_work_as_reviewer( self.student1 )
#         self.assertEqual( 2, len( work ), "Returns expected number of subs" )
#
#         for w in work:
#             self.assertIsInstance( w, Submission, "Returns expected type" )
#             if isinstance( w, InitialWork ):
#                 self.fail( "Returned submitter when looking for reviewer" )
#             else:
#                 self.assertEqual( w.reviewer.student_id, self.student1.student_id, "Review by expected student" )
#
#     def test_get_student_work( self ):
#         # check starting point
#         self.assertEqual( 6, len( self.obj.submissions ), "Starting with correct number" )
#
#         work = self.obj.get_student_work( self.student1 )
#         self.assertEqual( 5, len( work ), "Returns expected number of subs" )
#         for w in work:
#             self.assertIsInstance( w, Submission, "Returns expected type" )
#             if isinstance( w, InitialWork ):
#                 self.assertEqual( w.submitter, self.student1.student_id )
#
#     def test_calculate( self ):
#         self.obj.calculate()
#
#         self.assertEqual( len( self.obj.scores ), 2, "Number of scores matches number of students" )
#
#         for s in self.obj.scores:
#             self.assertIsInstance( s, UnitScores )
#             self.assertGreater( s.total, 0, "Total score greater than 0" )
#
#
# class TestAssignmentGradeCalculatorAbarrentCases( TestCase ):
#     def setUp( self ):
#         # Create the unit
#         self.assignment = unit_factory()
#
#         self.student1 = student_factory()
#         self.student2 = student_factory()
#
#         self.submissions = submissions_factory( self.student1, self.student2, self.assignment )
#
#         self.obj = UnitGradeCalculator( self.assignment, self.submissions )
#
#     def test_reviewer_flakes( self ):
#         """The student should receive full credit for the
#         unit if the reviewer doesn't do their part
#         """
#         pass
#
#     def test_missing_metareview( self ):
#         """The reviewer should receive full credit
#         if the original submitter does not complete the
#         metareview
#         """
#         pass



"""
Created by adam on 2/23/20
"""
from CanvasHacks.DAOs.sqlite_dao import SqliteDAO
from CanvasHacks.Logging.display import DisplayManager
from CanvasHacks.Repositories.reviewer_associations import AssociationRepository
from CanvasHacks.Repositories.students import StudentRepository

__author__ = 'adam'

from CanvasHacks import environment as env
import CanvasHacks.testglobals
if __name__ == '__main__':
    pass


class IStep:
    """Parent of all steps in a Skaa"""

    def __init__( self, course=None, unit=None, is_test=None, send=True ):
        """
        :param course:
        :param unit:
        :param is_test:
        :param send: Whether to actually send the messages
        """
        self.course = env.CONFIG.course if course is None else course
        self.unit = env.CONFIG.unit if unit is None else unit
        self.is_test = env.CONFIG.is_test if is_test is None else is_test
        self.send = send



    def _initialize( self ):
        """Creates and populates all relevant repositories and db access points"""
        self.studentRepo = StudentRepository( self.course )
        self.studentRepo.download()
        self._initialize_db()
        self.associationRepo = AssociationRepository( self.dao, self.activity_for_review_pairings)

        self.display_manager = DisplayManager(self.activity)

        # Moved to child classes while working on CAN-53
        # if isinstance(self.activity, MetaReview) and isinstance(self.activity_notifying_about, MetaReview):
        #     # If both of these are the metareview, then we are on the final
        #     # step (sending the metareview results to the reviewer). Thus
        #     # we need a special status repository since the notified field
        #     # for the metareview will have already been populated in the previous step
        #     self.statusRepos = FeedbackStatusRepository( self.dao, self.activity_notifying_about )
        # elif isinstance(self.activity, DiscussionReview) and isinstance(self.activity_notifying_about, DiscussionReview):
        #     # If both of these are the discussion review, then we are on the final
        #     # step (sending the review results to the poster). Thus
        #     # we need a special status repository since the notified field
        #     # for the review will have already been populated in the previous step
        #     self.statusRepos = FeedbackStatusRepository( self.dao, self.activity_notifying_about )
        # else:
        #     self.statusRepos = StatusRepository( self.dao, self.activity_notifying_about )

    def _initialize_db( self ):
        try:
            t = 'TEST-' if env.CONFIG.is_test else ""
            self.db_filepath = "{}/{}{}-Unit-{}-review-assigns.db".format( env.LOG_FOLDER, t, env.CONFIG.semester_name, self.unit.unit_number )
        except AttributeError as e:
            # This is likely to happen during testing
            print(e)

        if env.CONFIG.is_test:
            try:
                if CanvasHacks.testglobals.TEST_WITH_FILE_DB:
                    # testing: file db
                    self._initialize_file_db()
                    print( "Connected to TEST db file. {}".format( self.db_filepath ) )
                else:
                    # testing: in memory db
                    self._initialize_memory_db()
            except (NameError, AttributeError) as e:
                print(e)
                # The variable might not be defined under in any
                # number of circumstances. So default to the in-memory db
                self._initialize_memory_db()

        else:
            # real: file db
            self._initialize_file_db()
            print( "Connected to REAL db. {}".format( self.db_filepath ) )

    def _initialize_file_db( self ):
        self.dao = SqliteDAO( self.db_filepath )
        self.dao.initialize_db_file()

    def _initialize_memory_db( self ):
        self.dao = SqliteDAO()
        print( "Connected to in-memory testing db" )

    def run( self ):
        raise NotImplementedError



"""
Created by adam on 3/24/20
"""
__author__ = 'adam'

from CanvasHacks.Definitions.groups import ReviewType
from CanvasHacks.Definitions.journal import Journal
from CanvasHacks.GradingHandlers.assignment import AssignmentGrader
from CanvasHacks.GradingHandlers.journal import JournalGrader

if __name__ == '__main__':
    pass

# todo Switch to quiz_new when ready
from CanvasHacks.GradingHandlers.quiz import QuizGrader
from CanvasHacks.GradingHandlers.review import ReviewGrader


class GradingHandlerFactory:
    """
    Determines which sort of class which handles grading
    is needed and returns it
    """

    @staticmethod
    def make(  *args, **kwargs ):
        """
        Will likely contain
         work_repo=self.workRepo,
         submission_repo=self.subRepo,
         association_repo=self.association_repo

        :param args:
        :param kwargs:
        :return:
        """
        try:
            if isinstance(kwargs['activity'], ReviewType):
                return ReviewGrader(*args, **kwargs)

            if isinstance(kwargs['activity'], Journal):
                return JournalGrader(*args, **kwargs)

            if not kwargs['activity'].is_quiz_type:
                return AssignmentGrader(*args, **kwargs)

            return QuizGrader( *args, **kwargs )
        except KeyError:
            return QuizGrader(*args, **kwargs)




"""
Created by adam on 3/15/20
"""
__author__ = 'adam'

if __name__ == '__main__':
    pass


class InvalidGradingValuesError(Exception):
    """
    Raised when expecting a list of dictionaries of the
    format:
     { count : int,
            pct_credit : float
            }
    But got something bad
    """
    pass


"""
Created by adam on 2/23/20
"""
from CanvasHacks.Errors.messaging import NoStudentsNeedNotification
from CanvasHacks.Errors.review_pairings import NoReviewPairingsLoaded
from CanvasHacks.Logging.run_data import RunLogger
from CanvasHacks.Messaging.skaa import FeedbackFromMetareviewMessenger
from CanvasHacks.Repositories.factories import WorkRepositoryLoaderFactory
from CanvasHacks.Repositories.status import FeedbackStatusRepository, InvitationStatusRepository
from CanvasHacks.SkaaSteps.ISkaaSteps import IStep

__author__ = 'adam'


class SendMetareviewToReviewer( IStep ):

    def __init__( self, course=None, unit=None, is_test=None, send=True, **kwargs ):
        """
        :param course:
        :param unit:
        :param is_test:
        :param send: Whether to actually send the messages
        """
        super().__init__( course, unit, is_test, send, **kwargs )
        # The activity whose results we are going to be doing something with
        self.activity = unit.metareview

        # The activity which we are telling the student about
        # self.activity_notifying_about = unit.metareview

        # In sending the metareview results to the reviewer we are
        # telling the about the feedback on the review
        self.activity_feedback_on = unit.review

        # The activity_inviting_to_complete whose id is used to store review pairings for the whole SKAA
        self.activity_for_review_pairings = unit.initial_work

        self._initialize()

        self.feedback_status_repo = FeedbackStatusRepository(self.dao, self.activity_feedback_on, self.activity_for_review_pairings)

        self.statusRepos = [self.feedback_status_repo]

        self.associations = [ ]

    def run( self, **kwargs ):
        """
        Send feedback from the metareview to the person
        who completed the peer review
        :param only_new: Probably won't be used
        :param rest_timeout: Number of seconds to rest_timeout for canvas to generate report
        :return:
        """
        try:
            self._load_step( **kwargs )

            self._assign_step()

            self._message_step()

        except NoStudentsNeedNotification:
            print("All reviewers have already received metareview results")

    def _message_step( self ):
        print( "Going to send metareview results for {} students".format( len( self.associations ) ) )

        # Send
        self.messenger = FeedbackFromMetareviewMessenger( self.unit,
                                                          self.studentRepo,
                                                          self.work_repo,
                                                          self.statusRepos )
        self.messenger.notify( self.associations, self.send )
        # Log the run
        msg = "Sent {} metareview results to authors".format( len( self.associations ) )
        print( msg )

        lmsg = "{} \n {}".format( msg, self.associations )

        # Note we are distributing the material for the metareview, that's
        # why we're using that activity_inviting_to_complete.
        RunLogger.log_metareview_feedback_distributed( self.unit.metareview, lmsg )

    def _assign_step( self ):
        # Load the review pairs for all those in the work repo.
        # Some of these will have already been notified
        for student_id in self.work_repo.submitter_ids:
            # Work repo contains submitted meta reviews. Thus we look up
            # review pairings where a student submitting the metareview unit
            # is the assessee
            record = self.associationRepo.get_by_assessee( self.activity_for_review_pairings, student_id )
            if record is not None:
                self.associations.append( record )

        if len( self.associations ) == 0:
            raise NoReviewPairingsLoaded

        # Filter out students who have already been notified.
        self._filter_notified()


    def _load_step( self, **kwargs ):
        # Get work
        self.work_repo = WorkRepositoryLoaderFactory.make( self.activity, self.course, **kwargs )
        self.display_manager.initially_loaded = self.work_repo.data


    def _filter_notified( self ):
        """
        The downloaded store of student work contains metareviews completed
        by every author who has turned in the assignment.
        We need to remove authors who have already had their metareview responses sent to
        the person who had reviewed them.
        That's the job of this method.
        :return:
        """
        # Looking at assessor id because sending feedback from author to peer reviewer
        self.associations = [ ra for ra in self.associations if ra.assessor_id not in self.feedback_status_repo.previously_received_ids]

        if len(self.associations) == 0:
            # Everyone has already received their results
            raise NoStudentsNeedNotification

        # records = self.feedback_status_repo.reviewers_with_authors_sent_feedback
        # # We can remove those from the repo
        # self.work_repo.remove_student_records( records )
        # self.display_manager.post_filter = self.work_repo.data


if __name__ == '__main__':
    pass



"""
Created by adam on 3/14/20
"""
__author__ = 'adam'

from IPython.display import display
from ipywidgets import Layout, VBox, widgets

from CanvasHacks.StudentNaggers.discussion import DiscussionNagger
from CanvasHacks.StudentNaggers.discussion_review import DiscussionReviewNagger
from CanvasHacks.StudentNaggers.essay import EssayNagger
from CanvasHacks.StudentNaggers.skaa_review import SkaaReviewNagger

if __name__ == '__main__':
    pass


def nag_button( label, nag_obj, return_button=False, width='auto', **kwargs ):
    """
    :param nag_obj:
    :param label:
    :param return_button: Set to true if putting inside a box or something else which will call display
    :return:
    """

    RUNNING = False

    def get_style( is_running=False ):
        return 'warning' if is_running else 'danger'

    def get_name( is_running=False ):
        return 'RUNNING' if is_running else label

    # Create the button
    layout = widgets.Layout( width=width )
    b = widgets.Button( description=get_name( RUNNING ),
                        button_style=get_style( RUNNING ),
                        layout=layout )

    def callback( change ):
        RUNNING = True
        b.description = get_name( RUNNING )
        b.button_style = get_style( RUNNING )

        nag_obj.run()

        RUNNING = False

        b.description = get_name( RUNNING )
        b.button_style = get_style( RUNNING )

    b.on_click( callback )
    if return_button is True:
        # If putting inside a box or something else which will
        # call display
        return b
    else:
        display( b )


def nag_button_area( control_store, **kwargs ):
    # items_layout = Layout( width='auto' )  # override the default width of the button to 'auto' to let the button grow

    controls = [ ('ESSAY', EssayNagger( control_store[ 'skaa_repo' ] )),
                 ('SKAA REVIEW', SkaaReviewNagger( control_store[ 'skaa_repo' ] )),
                 ('DISCUSSION', DiscussionNagger( control_store[ 'diss_repo' ] )),
                 ('DISCUSSION REVIEW', DiscussionReviewNagger( control_store[ 'diss_repo' ] ))
                 ]

    box_layout = Layout(
        border='solid',
        **kwargs
    )
    buttons = [ widgets.Label( value="Nag students" ) ]
    buttons += [ nag_button( label, obj, return_button=True, **kwargs ) for label, obj in controls ]
    return VBox( buttons, layout=box_layout )



"""
Created by adam on 1/27/20
"""
from CanvasHacks.Processors.cleaners import TextCleaner
from CanvasHacks.Repositories.interfaces import IContentRepository
from CanvasHacks.Repositories.mixins import StudentWorkMixin
from CanvasHacks.Api.UploadGradeTools import upload_credit

__author__ = 'adam'

from CanvasHacks.Text.stats import WordCount


class DiscussionRepository( IContentRepository, StudentWorkMixin ):
    """Manages the data for one discussion unit"""

    def __init__( self, activity, course ):
        self.activity = activity
        self.topic_id = activity.topic_id
        self.course = course

        # List of dictionaries from parsed data:
        # [{'student_id', 'student_name', 'text'}]
        self.data = [ ]

        # The cleaner class that will be called to
        # remove html and other messy stuff from student
        # work
        self.text_cleaner = TextCleaner()

        self.analyzer = WordCount()


    @property
    def course_id( self ):
        return self.course.id

    def download( self ):
        # self._get_discussion_entries(topic_id)
        self._get_submissions( self.topic_id )
        self._parse_posts_from_submissions()
        print( "Loaded {} posts".format( len( self.data ) ) )

    def _get_submissions( self, topic_id ):
        """Retrieves all the information we'll need for grading
        Not using self.topic_id to allow method to be called
        independently for testing
        """
        topic = self.course.get_discussion_topic( topic_id )
        # Graded discussions will be tied to an unit, so
        # we need the id
        self.assignment_id = topic.assignment_id
        print( "Assignment {} is associated with topic {}".format( self.assignment_id, topic_id ) )
        # Load the unit object
        self.assignment = self.course.get_assignment( self.assignment_id )
        # Load all submissions for the unit
        self.submissions = { s.user_id: s for s in self.assignment.get_submissions() }
        print( "Loaded {} submissions for the unit".format( len( self.submissions.keys() ) ) )

    def _parse_posts_from_submissions( self ):
        """The submission objects downloaded for the unit will
        have the post information stored in a list called dicussion_entries.
        This takes all of those and loads the user id, name and text into
        posts"""
        for sid, submission in self.submissions.items():
            for entry in submission.discussion_entries:
                # Remove html and other artifacts from student answers
                # DO NOT UNCOMMENT UNTIL CAN-59 HAS BEEN FULLY TESTED
                content = self.text_cleaner.clean(entry[ 'message' ])

                self.data.append( { 'student_id': entry[ 'user_id' ],
                                    'student_name': entry[ 'user_name' ],
                                    'text': content
                                    } )
            #
            # self.posts.append( (entry.user_id, entry.user_name, entry.message) )

    def get_student_posts( self, student_id ):
        """Returns a list of all posts by student for the topic"""
        return [ p[ 'text' ] for p in self.data if p[ 'student_id' ] == student_id ]

        # return [ p.message for p in self.data if p.user_id == student_id ]

    def get_formatted_work_by( self, student_id ):
        """Returns all posts by the student, formatted for
        sending out for review or display"""
        posts = self.get_student_posts( student_id )
        # self._check_empty(posts)
        posts = "\n        -------        \n".join( posts )
        return posts

    def upload_student_grade( self, student_id, pct_credit ):
        upload_credit( self.course_id, self.assignment_id, student_id, pct_credit )
        # Not sure why this doesn't work, but doing it manually does
        # pct = "{}%".format(pct_credit) if isinstance(pct_credit, int) or pct_credit[-1:] != '%' else pct_credit
        # Look up the student submission
        # submission = self.submissions.get( student_id )
        # return submission.edit(posted_grade=pct)

    def display_for_grading( self ):
        """Returns student submissions in format expected for
        ipython display
        "Returns a list of dictionaries of all dicussion posts for the topic
        Format:
        """
        return self.data
        # [ e for e in self.data ]

    @property
    def submitter_ids( self ):
        """Returns a list of canvas ids of students who have submitted the unit"""
        # try:
        return list( set( [ s[ 'student_id' ] for s in self.data ] ) )

    # return [ (e.user_id, e.user_name, e.message) for e in self.data ]

    # @property
    # def student_ids( self ):
    #     uids = list( set( [ k['student_id'] for k in self.data ] ) )
    #     uids.sort()
    #     return uids

    @property
    def post_counts( self ):
        """Returns list of tuples
        ( student id, # of posts )
        """
        counts = [ ]
        for sid in self.student_ids:
            counts.append( (sid, len( [ s for s in self.data if s[ 'student_id' ] == sid ] )) )
        return counts

    def filter_by_count( self, min_post_count ):
        """
        Returns a copy of data without students who have not reached
        the minimum count
        :param min_post_count:
        :return:
        """
        students_to_keep = [ sid for sid, cnt in self.post_counts if cnt >= min_post_count ]
        return [ s for s in filter( lambda x: x[ 'student_id' ] in students_to_keep, self.data ) ]

    # def _get_discussion_entries(self, topic_id):
    #     """
    #     THIS WON'T WORK BECAUSE ONLY RETURNS TOP LEVEL ENTRIES
    #     Loads and returns a list of discussion objects.
    #         Objects look something like:
    #         {'id': 2132485, 'user_id': 169155,
    #         'parent_id': None, 'created_at': '2020-01-16T23:01:53Z',
    #         'updated_at': '2020-01-16T23:01:53Z', 'rating_count': None,
    #         'rating_sum': None, 'user_name': 'Test Student',
    #         'message': '<p>got em</p>', 'user': {'id': 169155,
    #         'display_name': 'Test Student',
    #         'avatar_image_url': 'https://canvas.csun.edu/images/messages/avatar-50.png',
    #         'html_url': 'https://canvas.csun.edu/courses/85210/users/169155',
    #         'pronouns': None, 'fake_student': True},
    #         'read_state': 'unread', 'forced_read_state': False,
    #         'discussion_id': 737847, 'course_id': 85210}
    #     """
    #     discussion = self.course.get_discussion_topic(topic_id)
    #     # result is lazy loaded, so iterate through it
    #     # self.data = [e for e in discussion.get_entries()]
    #     self.data = [e for e in discussion.get_topic_entries()]
    #     return self.data


if __name__ == '__main__':
    pass
